﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file HeightMap.hpp
 * @brief генератор случайного ландшафта, работает с картой высот (height map)
 * @author Кувшинов Д.Р. https://evetro.wordpress.com/
 */
#pragma once
#include <cstddef> // size_t
#include <cstdlib> // rand
#include <vector>


/// реализация карты высот, аккумулирует результаты разных генераторов высот
class HeightMap
{
public:
  /// тип чисел, используемый для представления значения высоты  
  using Height = float;
  using Row    = std::vector<Height>;
  using Size   = Row::size_type;

  /// создаёт пустую карту высот
  HeightMap() {}
  
  /// создать "плоскую" карту высот заданного размера rows x cols
  HeightMap(Size rows, Size cols)
    : data(rows)
  {
    for (auto &d: data)
      d.resize(cols);      
  }

  /// обменять содержимое с другой картой высот
  void swap(HeightMap &other)
  {
    data.swap(other.data);
  }

  /// сколько строк
  Size rows() const
  {
    return data.size();
  }

  /// сколько столбцов
  Size cols() const
  {
    return data.empty()? 0: data.front().size();
  }

  /// const-итератор начала заданной строки
  Row::const_iterator rowBegin(Size row) const
  {
    return data[row].begin();
  }

  /// const-итератор конца заданной строки
  Row::const_iterator rowEnd(Size row) const
  {
    return data[row].end();
  }

  /// получить высоту в заданных координатах (корректность координат не проверяется)
  Height operator()(Size row, Size col) const
  {
    return data[row][col];
  }

  /// ссылка на значение высоты в заданных координатах (корректность координат не проверяется)
  Height& operator()(Size row, Size col)
  {
    return data[row][col];
  }

  /// добавить к текущим значениям высот высоты, сгенерированные функтором bm типа BinMap (принимает координаты)
  template<class BinMap>
  void add(const BinMap &bm)
  {
    const Size rs = rows(), cs = cols();
    for (Size i = 0; i < rs; ++i)
    {
      auto &row = data[i];
      for (Size j = 0; j < cs; ++j)
        row[j] += bm(i, j);
    }
  }

private:
  std::vector<Row> data;
};



/// сгенерировать псевдослучайное число из отрезка [minv, maxv)
inline float random(float minv, float maxv)
{
  static const float anti_rand_max = 1.0f / float(RAND_MAX);
  return minv + (maxv - minv) * std::rand() * anti_rand_max;
}


/// генератор псевдослучайных высот с равномерным распределением
class RandomBinMap
{
  float from, to; // диапазон высот

public:
  /// установить диапазон высот
  explicit RandomBinMap(float from = 0.0f, float to = 1.0f)
    : from(from), to(to) {}

  /// получить следующее значение, координаты игнорируются
  float operator()(std::size_t, std::size_t) const
  {
    return random(from, to);
  }
};


/// базовый класс для классов -- генераторов "холмов"
class HillBase
{
protected:
  float cx, cy, // координаты центра холма
        h,      // высота
        rlen, rwid,       // "длина" и "ширина" холма (хранятся обратные величины)
        cos_phi, sin_phi; // косинус и синус угла поворота холма

  float transform_x(const float x, const float y) const
  {
    return rwid * (x * cos_phi - y * sin_phi);
  }

  float transform_y(const float x, const float y) const
  {
    return rlen * (x * sin_phi + y * cos_phi);
  }

public:
  /// инициализация состояния для заданных размеров карты и максимальной высоты
  /**
   * Для сокращения размера кода использует random.
   */
  void randomInit(float max_x, float max_y, float max_h)
  {
    cx = random(0.0f, max_x);
    cy = random(0.0f, max_y);
    h = random(max_h * 0.4f, max_h);
    rlen = 1.0f / (max_y * random(0.05f, 0.5f));
    rwid = 1.0f / (max_x * random(0.05f, 0.5f));
    cos_phi = random(-1.0f, 1.0f);
    sin_phi = 1.0f - cos_phi * cos_phi;
  }
};


/// колоколообразный холм
class Hill: public HillBase
{
public:
  /// вычислить высоту в заданных координатах
  float operator()(float x, float y) const
  {
    x -= cx; y -= cy; // центрировать
    // поворот и масштабирование
    const float nx = transform_x(x, y),
                ny = transform_y(x, y);
    // функция "колокол"
    return h / (1.0f + nx * nx + ny * ny);
  }

  float operator()(std::size_t i, std::size_t j) const
  {
    return (*this)(static_cast<float>(i), static_cast<float>(j));
  }
};


/// кольцевой холм (кратер)
class Ring: public HillBase
{
public:
  /// вычислить высоту в заданных координатах
  float operator()(float x, float y) const
  {
    x -= cx; y -= cy; // центрировать
    // поворот и масштабирование
    const float nx = transform_x(x, y),
                ny = transform_y(x, y);
    // сдвинутая функция "колокол"
    return h / (1.0f + std::abs(1.0f - nx * nx - ny * ny));
  }

  float operator()(std::size_t i, std::size_t j) const
  {
    return (*this)(static_cast<float>(i), static_cast<float>(j));
  }
};


/// генератор псевдослучайного ландшафта
inline void makeLandscape // inline, чтобы не создавать дополнительный .cpp файл
  (
    HeightMap &landscape ///< результирующий ландшафт (старый уничтожается)
  , std::size_t N        ///< количество строк карты высот - 1 
  , std::size_t M        ///< количество столбцов карты высот - 1
  , unsigned randoms     ///< сколько раз применить RandomBinMap
  , unsigned rings       ///< сколько раз применить Ring
  , unsigned hills       ///< сколько раз применить Hill
  )
{
  // создать карту высот
  HeightMap HM(N + 1, M + 1);
  const auto Nf = static_cast<float>(N),
             Mf = static_cast<float>(M);

  // применить RandomBinMap (делает карту бугристой)
  RandomBinMap rbm(-0.002f * (Nf + Mf), 0.001f * (Nf + Mf));
  while (randoms--)
    HM.add(rbm);

  // добавить кратеров
  Ring ring;
  while (rings--)
  {
    ring.randomInit(Nf, Mf, 0.75f * (Nf + Mf));
    HM.add(ring);
  }

  // добавить холмов
  Hill hill;
  while (hills--)
  {
    hill.randomInit(Nf, Mf, Nf + Mf);
    HM.add(hill);
  }

  // передать результат вызвавшему эту функцию коду
  landscape.swap(HM);
}
