﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file TerrainGraph.hpp
 * @brief реализация графа ландшафта (длины рёбер вычисляются с помощью карты высот), на котором испытываются алгоритмы Дейкстры и A*
 * @author Кувшинов Д.Р. https://evetro.wordpress.com/
 */
#pragma once
#include <limits>
#include <vector>
#include <functional>
#include <algorithm>

#include <avo/Vec3.hpp>
#include <avo/Color.hpp>
#include <avo/BoundedArray.hpp>

#include "HeightMap.hpp"


/// вершина с заданными координатами, нормалью и цветом
struct Vertex
{
  avo::Color c;
  avo::Vec3f p, n;
};


/// ландшафт (terrain) из треугольников, реализация отрисовки
class Terrain
{
  std::vector<Vertex> data; // массив вершин, std::vector гарантирует размещение элементов в массиве
  std::size_t rs, cs; // размеры ландшафта

public:
  using Size = std::size_t;

  /// создать пустой
  Terrain()
    : rs(0), cs(0) {}

  /// разметить массив для ландшафта заданного размера
  Terrain(const Size rows, const Size cols)
    : data(2u * (rows + 1) * cols), rs(rows), cs(cols) {}

  /// создать ландшафт из карты высот с заданными шагами по 0x и по 0y
  explicit Terrain(const HeightMap &hm, const float dx = 1.f, const float dy = 1.f)
    : data(2u * hm.rows() * hm.cols())
    , rs(hm.rows() - 1u)
    , cs(hm.cols())
  {
    fromHeightMap(hm, dx, dy);
  }

  /// заполнить массив вершин по заданной карте высот
  /**
   * @param dx шаг клетки по 0x
   * @param dy шаг клетки по 0y
   */
  void fromHeightMap(const HeightMap &HM, const float dx, const float dy)
  {
    for (Size i = 0, rows = HM.rows(); i < rows; ++i)
      prepareRow(i, dx, dy, HM.rowBegin(i), HM.rowEnd(i));
    finishRows();
  }


  /// количество строк (полосок треугольников)
  Size rows() const
  {
    return rs;
  }

  /// количество столбцов (размер полоски треугольников пополам)
  Size cols() const
  {
    return cs;
  }


  /// извлечь координаты вершины в заданных координатах
  /**
   * Значение row == rows() соответствует последней строке в исходной карте высот.
   */
  const avo::Vec3f& point(Size row, Size col) const
  {
    const Size final = static_cast<Size>(row == rs),
               index = 2 * ((row - final) * cs + col) + final;
    return data[index].p;
  }

  /// извлечь цвет треугольника в заданных координатах (полоска, номер в полоске)
  avo::Color& color(Size strip, Size num)
  {
    return data[2 * strip * cs + num + 2].c;
  }

  /// получить цвет треугольника по номеру вершины графа ландшафта (вершины графа отвечают треугольникам модели)
  avo::Color& color(Size index)
  {
    const Size c2 = 2 * (cs - 1),
               idx = (index / c2) * (2 * cs) + index % c2 + 2;
    return data[idx].c;
  }


  /// нарисовать ландшафт
  void paint() const
  {
    if (data.empty())
      return; // нечего рисовать

    const Size c2 = 2u * cs;
    // подключить массивы
    glColorPointer  (4, GL_FLOAT, sizeof(Vertex), &(data[0].c));
    glNormalPointer (   GL_FLOAT, sizeof(Vertex), &(data[0].n));
    glVertexPointer (3, GL_FLOAT, sizeof(Vertex), &(data[0].p));

    // нарисовать (одна полоска за вызов)
    for (Size i = 0, offset = 0; i < rs; ++i, offset += c2)
      glDrawArrays(GL_TRIANGLE_STRIP, offset, c2);

    // рисование массивов работает аналогично следующему коду (но, вероятно, быстрее):
    /*
      for (Size i = 0; i < rs; ++i)
      {
        glBegin(GL_TRIANGLE_STRIP);
        for (Size j = 0; j < c2; ++j)
        {
          const Vertex &p = data[i * c2 + j];
          p.c.select();
          glNormal3f(p.n.x, p.n.y, p.n.z);
          glVertex3f(p.p.x, p.p.y, p.p.z);
        }
        glEnd();
      }
    */
  } 

private:

  /// заполнить координаты точек строки
  template <class VecIt> // VecIt -- итератор строки карты высот (координаты по 0z)
  void prepareRow(Size row, float dx, float dy, VecIt begin, VecIt end)
  {
    if (begin == end || row > rs) return;
 
    const Size final = row == rs? 1: 0;
    const Size c2 = 2u * cs;

    // позиции
    auto p = data.begin() + (row - final) * c2;
    for (Size i = final; i < c2; i += 2u)
    {
      const avo::Vec3f pos((i / 2) * dx, row * dy, *begin++);
      p[i].p = pos;
      if (begin == end)
        break;
    }

    // цвета и нормали "по умолчанию"
    const avo::Vec3f default_n(0.0f, 0.0f, -1.0f);
    for (Size i = 0; i < c2; ++i)
    {
      p[i].n = default_n;
      p[i].c = avo::WHITE;
    }
  }


  /// вычислить нормали для корректной работы повершинного освещения
  void finishRows()
  {
    const Size c2 = 2u * cs;
    const avo::Vec3f default_n(0.f, 0.f, 1.f);
    for (Size i = 0; i < rs - 1u; ++i)
    {
      auto p = data.begin() + i * c2,
           q = p + c2;

      for (Size j = 1; j < c2; j += 2u)
      {
        p[j].p = q[j - 1u].p;
        
        // нормали
        const auto dp = p[j + 1u].p - p[j].p;
        auto n1 = cross(p[j + 1u].p - p[j - 1u].p, dp);
        auto n2 = cross(q[j + 1u].p - p[  j   ].p, dp);      
        n1.normalize();
        n2.normalize();

        p[j + 1u].n = n1;
        p[j + 2u].n = n2;
      }
    }

    // нормали последней строки вершин
    auto p = data.begin() + (rs - 1u) * c2;
    for (Size j = 0; j < c2 - 2u; j += 2u)
    {
      auto n = cross(p[j + 1u].p - p[j].p, p[j + 2u].p - p[j].p);
      n.normalize();
      p[j + 2u].n = n;
    }
    
    for (Size j = 1; j < c2 - 2u; j += 2u)
    {
      auto n = cross(p[j + 2u].p - p[j].p, p[j + 1u].p - p[j].p);
      n.normalize();
      p[j + 2u].n = n;
    }
  }
};


/// граф, отвечающий отношению соседства между треугольниками, из которых формируется ландшафт (позиции размещаются в центрах треугольников)
/**
 * У каждого треугольника до четырёх соседей: треугольники, соседние через ребро,
 * и один треугольник, соседний через вершину (при прямом угле в проекции).
 */
class TerrainGraph
{
public:
  using VertexIndex = std::size_t;
  using Adjacency   = avo::BoundedArray<VertexIndex, 4>; ///< список соседей
  using value_type  = Adjacency;

private:
  using Vertices    = std::vector<avo::Vec3f>; // данные вершин
  using Adjacencies = std::vector<Adjacency>;  // собственно граф (массив списков соседей)

  Vertices V;
  Adjacencies E;

public:

  TerrainGraph() {}

  /// инициализировать граф в соответствии с ландшафтом
  explicit TerrainGraph(const Terrain &T)
    : V(2 * T.rows() * (T.cols() - 1))
    , E(V.size())
  {
    fromTerrain(T);
  }


  /// количество вершин в графе (треугольников в ландшафте)
  VertexIndex vertices() const
  {
    return V.size();
  }

  /// оценка стоимости (длины) перехода из вершины u в вершину v (по прямой в пространстве ландшафта)
  float estimatePrice(VertexIndex u, VertexIndex v) const
  {
    // некая "хитрая" функция, которая учитывает расстояние в проекции и крутость перехода
    if (u == v)
      return 0.0f;
    
    avo::Vec3f d = V[v] - V[u];
    const float dist = d.length(), dz = d.z;
    if (dist == 0.0f)
      return 0.0f;

    const float xy_dist = std::hypot(d.x, d.y); // длина в проекции на плоскость 0xy
    const float x = 4.0f * dz / xy_dist, x2 = x * x;
    return dist + x2;
  }

  /// максимальная оценка стоимости перехода в заданную вершину to_v
  float maxPrice(VertexIndex to_v) const
  {
    float maxprice = 0.0f;
    for (VertexIndex i = 0, verts = vertices(); i < verts; ++i)
    {
      const float p = estimatePrice(i, to_v);
      if (maxprice < p)
        maxprice = p;
    }

    return maxprice;
  }


  /// получить длину ребра (u, v), возвращает бесконечность для несуществующих рёбер
  float weight(VertexIndex u, VertexIndex v) const
  {
    // так как списки смежностей неупорядочены и имеют небольшой размер,
    // разумно использовать прямой линейный поиск для проверки наличия искомого значения
    if (std::find(E[u].begin(), E[u].end(), v) != E[u].end())
      return estimatePrice(u, v);
    // не нашли
    return std::numeric_limits<float>::infinity();
  }

  
  /// получить список соседей заданной вершины
  const Adjacency& operator[](VertexIndex v) const
  {
    return E[v];
  }


  /// нарисовать рёбра графа, используя сохранённые позиции вершин в пространстве ландшафта
  void paint() const
  {
    glBegin(GL_LINES);
    glNormal3f(0.0f, 0.0f, 1.0f);
    for (VertexIndex i = 0, Esz = E.size(); i < Esz; ++i)
    {
      const auto &Ei = E[i];
      for (VertexIndex j = 0, Eisz = Ei.size(); j < Eisz; ++j)
      {
        const auto &Vi = V[i];
        const auto &VEij = V[Ei[j]];

        glVertex3f(Vi.x, Vi.y, Vi.z);
        glVertex3f(VEij.x, VEij.y, VEij.z);
      }
    }
    glEnd();
  }

  /// нарисовать путь, ведущий в вершину s, в соответствии с позициями вершин в пространстве ландшафта
  /**
   * Путь path представляет собой массив предшественников, если path[v] == v, то v считается началом пути.
   */
  template <class Path>
  void paint(const Path &path, const VertexIndex s) const
  {
    glBegin(GL_LINE_STRIP);
    glNormal3f(0.0f, 0.0f, 1.0f);
    VertexIndex p = s;
    while (true) // более безопасный вариант: for (VertexIndex i = 0, verts = vertices(); i < verts; ++i)
    {
      const VertexIndex prev = path[p];
      if (prev == p)
        break;

      const auto &Vp = V[p];
      const auto &Vprev = V[prev];
      p = path[p];

      glVertex3f(Vp.x, Vp.y, Vp.z);
      glVertex3f(Vprev.x, Vprev.y, Vprev.z);
    }
    glEnd();
  }


  /// сгенерировать граф, отвечающий треугольникам ландшафта T
  void fromTerrain(const Terrain &T)
  {
    V.clear();
    E.clear();

    const VertexIndex Vsz  = 2 * T.rows() * (T.cols() - 1),
                      Vsz1 = Vsz - 1;
    V.resize(Vsz);
    E.resize(Vsz);

    const VertexIndex rs = T.rows(), cs = T.cols() - 1;
    for (VertexIndex i = 0; i < rs; ++i)
    {
      for (VertexIndex j = 0; j < cs; ++j)
      {
        // координаты
        const VertexIndex k = 2 * (i * cs + j);
        auto &Vk  = V[k];
        auto &Vk1 = V[k + 1];
        Vk  = tricenter(T.point(i, j), T.point(i + 1, j), T.point(i, j + 1));
        Vk1 = tricenter(T.point(i + 1, j + 1), T.point(i + 1, j), T.point(i, j + 1));
        Vk.z  += 1.0f;
        Vk1.z += 1.0f;
      }
    }

    // списки соседей
    const VertexIndex c2 = 2u * cs;
    // углы
    {
      E[0].resize(1);
      E[0].front() = 1;

      auto &Ev = E[Vsz1];
      Ev.resize(1);
      Ev.front() = Vsz - 2;
    }
    
    // "нижняя сторона"
    for (VertexIndex i = 2; i < c2; i += 2u)
    {
      auto &Ei1 = E[i - 1];
      Ei1.resize(4);
      Ei1[0] = i - 2u;
      Ei1[1] = i;
      Ei1[2] = i - 2u + c2;
      Ei1[3] = i + c2;

      auto &Ei = E[i];
      Ei.resize(2);
      Ei.front() = i - 1u;
      Ei.back()  = i + 1u;
    }

    // "верхняя сторона"
    for (VertexIndex i = Vsz - c2 + 1; i < Vsz1; i += 2u)
    {
      auto &Ei1 = E[i + 1];
      Ei1.resize(4);
      Ei1[0] = i;
      Ei1[1] = i + 2u;
      Ei1[2] = i - c2;
      Ei1[3] = i + 2u - c2;

      auto &Ei = E[i];
      Ei.resize(2);
      Ei.front() = i - 1u;
      Ei.back()  = i + 1u;
    }

    // "левая сторона"
    for (VertexIndex i = c2; i < Vsz; i += c2)
    {
      const VertexIndex k = i - c2 + 1u;
      auto &Ek = E[k];
      if (Ek.empty())
      {
        Ek.resize(4);
        Ek[0] = i;
        Ek[1] = k - 1u;
        Ek[2] = k + 1u;
        Ek[3] = i + 2u;
      }

      auto &Ei = E[i];
      Ei.resize(2);
      Ei.front() = k;
      Ei.back()  = i + 1u;
    }

    // "правая сторона"
    for (VertexIndex i = c2 - 1; i < V.size() - 1; i += c2)
    {
      const VertexIndex k = i + c2 - 1u;
      auto &Ek = E[k];
      if (Ek.empty())
      {
        Ek.resize(4);
        Ek[0] = i;
        Ek[1] = k - 1u;
        Ek[2] = k + 1u;
        Ek[3] = i - 2u;
      }        

      auto &Ei = E[i];
      Ei.resize(2);
      Ei.front() = i - 1u;
      Ei.back()  = k;
    }

    // средняя часть
    for (VertexIndex i = 1; i < rs - 1; ++i)
    {
      for (VertexIndex j = 1; j < cs - 1; ++j)
      {
        const VertexIndex k = i * c2 + 2 * j;
        auto &Ek  = E[k];
        auto &Ek1 = E[k + 1];
        Ek.resize(4);
        Ek[0] = k - 1u;
        Ek[1] = k + 1u;
        Ek[2] = k - 1u - c2;
        Ek[3] = k + 1u - c2;

        Ek1.resize(4);
        Ek1[0] = k; 
        Ek1[1] = k + 2u;
        Ek1[2] = k + c2;
        Ek1[3] = k + c2 + 2u;
      }
    }
  }
};


/// класс-адаптер, оборачивающий TerrainGraph::price в качестве эвристики для A*
class HeuristicForGraph
{
  const TerrainGraph &g;
  const float factor;

public:
  HeuristicForGraph(const TerrainGraph &g, const float factor)
    : g(g), factor(factor) {}

  float operator()(std::size_t u, std::size_t v) const
  {
    return factor * g.estimatePrice(u, v);
  }
};

/// класс-адаптер, представляющий TerrainGraph::weight в виде матрицы весов
class WeightsForGraph
{
  const TerrainGraph &g;

public:
  using value_type = float;

  WeightsForGraph(const TerrainGraph &g) : g(g) {}

  value_type operator()(std::size_t u, std::size_t v) const
  {
    return g.weight(u, v);
  }
};
