﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: GLSL Plasma
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, GLEW (http://glew.sourceforge.net/basic.html), OpenGL 2.0.
 * Основан на примерах  http://en.wikibooks.org/wiki/OpenGL_Programming и AV Orchid/Plasma
 * Требует наличия plasma.glfs с текстом шейдера рядом с .exe.
 *
 * Эффект "плазмы". См. http://en.wikipedia.org/wiki/Plasma_effect
 *
 * Нажатие:
 * ESC/Q -- выход из программы;
 * цифры -- выбор файла пиксельного шейдера (0 -- 9.glsl.frag)
 */
#include <GL/glew.h>
#include <GL/freeglut.h>

#include <iostream> // использует cout для вывода FPS
#include <fstream>
#include <string>
#include <utility>
#include <cstdint>
#include <cctype>
#include <algorithm>

#include <avo/Timer.hpp>


//
// константы и глобальные переменные
//

/// имя файла пиксельного шейдера
/**
 * Подразумевает подмену первого символа.
 * Не слишком элегантное и масштабируемое решение, зато весьма простое :)
 * Предполагается, что шейдеров не может быть больше 10 (они привязаны к цифровым клавишам).
 */
char fragment_shader_filename[] = "0.glsl.frag\0";

/// счётчик кадров (сбрасывается каждые 10с)
int frames;
/// момент времени, с которого происходит накопление счётчика кадров
double frames_reported_last;
/// момент начала рендеринга последнего кадра
double last_t;

// индексы
GLuint program = -1,    ///< индекс программы (вершинный + пиксельный шейдеры), 
       vbuffer,         ///< индекс буфера вершин (прямоугольник на весь экран),
       palette_texture; ///< индекс текстуры, используемой как палитра

// индексы параметров шейдеров
GLint attribute_coord2d; ///< атрибут вершинного шейдера
GLint uniform_palette,   ///< семплер текстуры палитры (не используется)
      uniform_t,         ///< параметр "время" пиксельного шейдера
      uniform_pixel_units_coef, ///< параметр масштаба пиксельного шейдера
      uniform_width,     ///< параметр для передачи ширины окна в пиксельный шейдер
      uniform_height;    ///< параметр для передачи высоты окна в пиксельный шейдер

// текущие значения параметров, передаваемые пиксельному шейдеру (время и размеры окна в пикселях)
float t, width = 800.f, height = 600.f;

/// поток вывода для сообщений об ошибках (файл)
std::ofstream err;


//
// вспомогательные функции
// осторожно: неявно используют err для сообщений об ошибках

/// прочитать содержимое файла с заданным именем в строку
std::string file_to_string(const char *fn)
{
  // попробуем открыть нужный файл
  std::ifstream input_file(fn, std::ios::binary);
  if (!input_file) // ошибка, файл открыть не удалось
  {
    err << "File \"" << fn << "\" can not be opened.\n";
    return {};
  }

  // установить позицию чтения файла в конец файла
  input_file.seekg(0, std::ios::end);
  // получить численное значение позиции чтения (т.е. размер файла в байтах)
  const std::streamsize input_file_size = input_file.tellg();

  // переменная, в которую будем считывать содержимое файла
  std::string result;
  if (input_file_size >= result.max_size() - 1) // ошибка, файл слишком велик
  {
    err << "File \"" << fn << "\" is too big (" << input_file_size << " bytes).\n";
    return {};
  }

  // подготовить строку нужного размера
  result.resize(static_cast<std::size_t>(input_file_size) + 1, '\0');
  // установить позицию чтения файла в начало
  input_file.seekg(std::ios::beg);
  // считать содержимое в строку (в подготовленное для этого место)
  input_file.read(&result[0], input_file_size);
  return result; // перемещение без копирования в C++11
}


// получить индекс параметра шейдера, в качестве prog передаётся program
bool bind_uniform(GLuint prog, GLint &loc, const char *name)
{
  loc = glGetUniformLocation(prog, name);
  if (loc >= 0)
    return true;

  err << "Could not bind uniform \"" << name << "\".\n";
  return false;
}

// получить индекс атрибута шейдера
bool bind_attribute(GLuint prog, GLint &loc, const char *name)
{
  loc = glGetAttribLocation(program, "coord2d");
  if (loc >= 0)
    return true;

  err << "Could not bind attribute \"" << name << "\".\n";
  return false;
}


/// вывести сообщения об ошибках компиляции шейдера (компилятор входит в состав драйвера ГП) или компоновки программы
void log_errors(GLuint object)
{
  // получить длину текста сообщений
  GLint log_length = 0;
  if (glIsShader(object))
    glGetShaderiv(object, GL_INFO_LOG_LENGTH, &log_length);
  else if (glIsProgram(object))
    glGetProgramiv(object, GL_INFO_LOG_LENGTH, &log_length);
  else
    return (void)(err << "Checking shader: not a shader or a program.\n");

  if (log_length == 0)
    return;

  // подготовить место для текста
  std::string log(log_length, '\0');

  // получить текст сообщений
  if (glIsShader(object))
    glGetShaderInfoLog(object, log_length, NULL, &log[0]);
  else if (glIsProgram(object))
    glGetProgramInfoLog(object, log_length, NULL, &log[0]);

  // вывести сообщения в поток сообщений об ошибках
  err << "Shader errors:\n";
  err << log << '\n';
}


//
// управление ресурсами
//


/// отправить шейдерам значения параметров
void update_program_parameters()
{
  glUniform1f(uniform_pixel_units_coef, 8.f / std::max(width, height));
  glUniform1f(uniform_width, width);
  glUniform1f(uniform_height, height);
}


/// инициализация шейдеров и сборка программы (вершинный + фрагментный==пиксельный шейдеры)
bool init_program()
{
  if (program != -1)
    glDeleteProgram(program);

  GLint compile_ok = GL_FALSE,
        link_ok    = GL_FALSE;
 
  // вершинный шейдер: тривиальный
  GLuint vs = glCreateShader(GL_VERTEX_SHADER);
  const char *vs_source = 
    "#version 110\n"  // OpenGL 2.0
    "attribute vec2 coord2d;                  "
    "void main(void) {                        "
    "  gl_Position = vec4(coord2d, 0.0, 1.0); "
    "}";
  glShaderSource(vs, 1, &vs_source, NULL);
  glCompileShader(vs);
  glGetShaderiv(vs, GL_COMPILE_STATUS, &compile_ok);
  if (!compile_ok)
  {
    err << "Can't compile the vertex shader.\n";
    log_errors(vs);
    return false;
  }

  // пиксельный (фрагментный) шейдер: загружается из файла
  std::string fs_data = file_to_string(fragment_shader_filename);
  if (fs_data.empty())
  {
    err << fragment_shader_filename << " (fragment shader source) not found or is empty\n";
    return false;
  }
  const char *fs_source = fs_data.c_str(); // glShaderSource takes char* array

  GLuint fs = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(fs, 1, &fs_source, NULL);
  glCompileShader(fs);
  glGetShaderiv(fs, GL_COMPILE_STATUS, &compile_ok);
  if (!compile_ok)
  {
    err << "Can't compile the fragment shader.\n";
    log_errors(fs);
    return false;
  }

  // компоновка программы (комбинация вершинного и пиксельного шейдеров)
  program = glCreateProgram();
  glAttachShader(program, vs);
  glAttachShader(program, fs);
  glLinkProgram(program);
  glGetProgramiv(program, GL_LINK_STATUS, &link_ok);
  if (!link_ok)
  {
    err << "Can't link shaders.\n";
    log_errors(program);
    return false;
  }

  // получить индексы атрибутов вершинного шейдера
  if (!bind_attribute(program, attribute_coord2d, "coord2d"))
    return false;

  // получить индексы параметров пиксельного шейдера
  if (!(bind_uniform(program, uniform_palette, "palette")
     && bind_uniform(program, uniform_t, "t")
     && bind_uniform(program, uniform_pixel_units_coef, "pixel_units_coef")
     && bind_uniform(program, uniform_width, "width")
     && bind_uniform(program, uniform_height, "height")))
    return false;

  // включить скомпонованную программу
  glUseProgram(program);
  update_program_parameters();
  return true;
}


/// инициализация используемых ресурсов (шейдеры, буферы, текстуры)
bool init_resources()
{
  if (!GLEW_VERSION_2_0)
  {
    err << "Sorry, we need at least OpenGL 2.0 support! You have " << glGetString(GL_VERSION) << ".\n";
    return false;
  } 

  // шейдеры
  if (!init_program())
    return false;

  // вершинный буфер

  // выделить новый индекс для буфера и включить этот буфер
  glGenBuffers(1, &vbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, vbuffer);
  // будем рисовать прямоугольник на весь экран
  GLfloat vertices[] = 
  {
     -1.f,  -1.f,
      1.f,  -1.f,
     -1.f,   1.f,
      1.f,   1.f
  };

  // передать данные и включить режим рисования из буфера, который больше не будет меняться (static)
  glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
  glEnableVertexAttribArray(attribute_coord2d);

  glVertexAttribPointer(
    attribute_coord2d, // наш атрибут -- это пара координат x, y
    2,                 // количество элементов на вершины (две координаты)
    GL_FLOAT,          // тип элемента
    GL_FALSE,          // не включать автоматическую нормализацию
    0,                 // между группами данных нет пропусков
    (void*)(0)         // сдвиг начала данных в буфере равен нулю (используем буфер целиком)
  ); // приведение к указателю по историческим причинам (в OpenGL 1.x передавали указатель на массив в системной памяти)


  // текстуры

  // создадим палитру, отвечающую цветам http://ru.wikipedia.org/wiki/HSV_(цветовая_модель) ,
  // в виде одномерной текстуры, воспользуемся возможностями аппаратной линейной интерполяции, предоставляемой ГП
  glGenTextures(1, &palette_texture);
  glBindTexture(GL_TEXTURE_1D, palette_texture);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  // семь цветовых значений на основе базовых 6 цветов, 
  // ГП должен поддерживать текстуры размера не степени 2 (есть при поддержке OpenGL 2.0 и новее)
  std::uint32_t palette[] =
  {
    0x0000FF, // red
    0xFF00FF, // magenta
    0xFF0000, // blue
    0xFFFF00, // cyan
    0x00FF00, // green
    0x00FFFF, // yellow
    0x0000FF  // red
  };

  // передать palette в качестве содержимого текущей текстуры
  glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA, 7, 0, GL_RGBA, GL_UNSIGNED_BYTE, palette);

  // привязать текущую текстуру к семплеру 0
  glActiveTexture(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_1D, palette_texture);

  // считаем инициализацию успешной
  return true;
}

/// удаление выделенных ресурсов
void free_resources()
{
  glDeleteTextures(1, &palette_texture); // удалить текстуру (палитру)
  glDeleteBuffers(1, &vbuffer); // удалить вершинный буфер (прямоугольник)
  glDeleteProgram(program); // удалить шейдеры
}


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{
  // получить текущее время
  const double current_t = avo::seconds();
  if (current_t == last_t)
    return;

  const float delta_t = static_cast<float>(current_t - last_t);
  last_t = current_t;
  t = t + delta_t == t? 0.f: t + delta_t; // сбрасывать счётчик t, если он перестаёт обновляться

  // передать текущее время шейдеру
  glUniform1f(uniform_t, t);
  // нарисовать наш прямоугольник во весь экран
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

  // счётчик кадров в секунду, выводит среднее за 10с
  ++frames;
  const double frames_delta_t = current_t - frames_reported_last;
  if (frames_delta_t >= 10.0)
  {
    std::cout << frames / frames_delta_t << '\n';
    frames = 0;
    frames_reported_last += frames_delta_t;
  }

  // показать результат на экране
  glFinish();
  glutSwapBuffers();
}


/// изменили размеры окна, нужно передать новые значения параметров в шейдер
void reshape(int w, int h)
{
  glViewport(0, 0, w, h);
  width   = static_cast<float>(w);
  height  = static_cast<float>(h);
  update_program_parameters();
}


/// обработка нажатий клавиш
void handle_key(unsigned char key, int x, int y)
{
  switch(key) // код клавиши
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  // сбросить время
  case 't': case 'T': case 'е': case 'Е':
    last_t = avo::seconds();
    t = 0.0f;
    break;

  default:
    if (std::isdigit(static_cast<unsigned char>(key)))
    {
      const char old_file = fragment_shader_filename[0];
      fragment_shader_filename[0] = key;
      // перезагрузить шейдер, пересобрать программу
      if (!init_program())
      {
        // если не удалось, вернуться к старому шейдеру
        fragment_shader_filename[0] = old_file;
        init_program();
      }
    }
  }
}


/// вся анимация "спрятана" в функции перерисовки
void post_redisplay()
{
  glutPostRedisplay();
}


//
// main
//
 
int main(int argc, char* argv[])
{
  // инициализация GLUT (создание окна)
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowSize(800, 600);
  glutCreateWindow("GLSL Plasma");
  err.open("GLSLPlasma.log");
 
  // инициализация GLEW (узнать возможности используемой реализации OpenGL)
  GLenum glew_status = glewInit();
  if (glew_status != GLEW_OK)
  {
    err << "GLEW error: " << glewGetErrorString(glew_status) << ".\n";
    err.close();
    return EXIT_FAILURE;
  }

  // инициализация ресурсов, привязка обработчиков событий и старт цикла обработки сообщений
  if (init_resources())
  {
    glutDisplayFunc(repaint);     // обработчик перерисовки
    glutIdleFunc(post_redisplay); // постоянно вызывается в цикле
    glutKeyboardFunc(handle_key); // обработчик нажатий клавиш
    glutReshapeFunc(reshape);     // обработчик изменения размеров окна
    glutMainLoop();               // запуск цикла обработки сообщений
    free_resources();             // освободить ресурсы
    err.close();
    return EXIT_SUCCESS;          // успешное завершение работы
  }

  // инициализация ресурсов не удалась
  err << "Initialization of GL resources failed.\n";
  err.close();
  return EXIT_FAILURE;
}
