﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: GLUPoly
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Демонстрация использования GLU для триангуляции многоугольников.
 * Управление "камерой" -- стрелки и +/- для увеличения/уменьшения масштаба.
 * HOME -- вернуть камеру в положение по умолчанию.
 * SPACE/ENTER -- загрузка файла (название надо ввести в консоли).
 * ESC -- выход из программы.
 */

#include "Contour.hpp"
#include "Mesh2.hpp"

#include <avo/Camera2.hpp>
#include <avo/Timer.hpp>

#include <cmath>
#include <string>
#include <memory>
#include <functional>
#include <algorithm>
#include <iostream>
#include <fstream>

using avo::Color;
using avo::Camera2;


/// определение "сцены"
class Scene
{
  Mesh2 mesh; // сетка, представляющая многоугольник
  Bbox2 mbox; // бокс, в который вписана сетка (для автопозиционирования камеры)

public:

  /// загрузить набор контуров из файла fn
  void load(const std::string &fn)
  {
    std::ifstream f(fn);
    if (!f)
      std::cout << "File \"" << fn << "\" can't be loaded." << std::endl;
    else
      load(f);
  }

  /// считать набор контуров из потока
  void load(std::istream &is)
  {
    using namespace std;

    int n = 0; // количество контуров
    is >> n;

    if (n > 0)
    {
      cout << n << " contours.\n";
      vector<Contour> ctrs(n);

      // прочитать из файла первый контур и сформировать бокс
      is >> ctrs[0];
      auto   verts = ctrs[0].size();
      Scalar area  = ctrs[0].area(),
             perim = ctrs[0].perimeter();
      mbox = ctrs[0].box();

      // прочитать оставшиеся контуры
      for (int i = 1; i < n && is.good(); ++i)
      {
        is >> ctrs[i];
        verts += ctrs[i].size();
        area  += ctrs[i].area();
        perim += ctrs[i].perimeter();
        mbox  += ctrs[i].box();
      }

      // сводка по загруженным контурам
      cout << verts << " vertices, area = " << area << ", perimeter = " << perim;
      cout << "\nbox: min = " << mbox.pmin << "; max = " << mbox.pmax << endl;

      // ликвидировать пустые контуры
      ctrs.erase
        (
          remove_if
            (
              ctrs.begin(), ctrs.end(), // mem_fun не хочет работать, так пусть будет лямбда :)
              [](const Contour &c) { return c.empty(); } // std::mem_fn(&Contour::empty)
            ),
          ctrs.end()
        );

      // построить сетку
      try
      {
        mesh = Mesh2(ctrs.begin(), ctrs.end());
        mesh.background.alpha = 0.4f;
      }
      catch (...)
      {
        cout << "Failed to create new mesh." << endl;
      }
    }
    else
      cout << "No contours found." << endl;
  }


  /// изобразить "сцену"
  void paint() const
  {
    mesh.paint();
  }

  /// получить бокс
  const Bbox2& box() const
  {
    return mbox;
  }
};


//
// глобальные объекты

Scene scene;
Camera2 camera;
 

//
// колбеки FreeGlut

/// перерисовка
void repaint()
{
  glClear (GL_COLOR_BUFFER_BIT);

  // координатная сетка
  const Bbox2 &box = camera.box();
  const Scalar 
    h = box.height(),
    w = box.width(),
    scale   = std::floor(std::log10(std::min(h, w))),
    step    = std::pow(10.0, scale),
    invstep = 1.0 / step
  ;

  glBegin (GL_LINES);
  {
    const double
      xbeg = step * std::ceil(box.pmin.x * invstep),
      xend = step * std::ceil(box.pmax.x * invstep)
    ;

    Color(avo::GREY, 0.5f).select ();
    // вертикальные
    for (double x = xbeg; x < xend; x += step)
    {
      glVertex2d (x, box.pmin.y);
      glVertex2d (x, box.pmax.y);
    }

    const double
      ybeg = step * std::ceil(box.pmin.y * invstep),
      yend = step * std::ceil(box.pmax.y * invstep)
    ;

    // горизонтальные
    for (double y = ybeg; y < yend; y += step)
    {
      glVertex2d (box.pmin.x, y);
      glVertex2d (box.pmax.x, y);
    }

    // оси
    avo::WHITE.select();
    if (box.pmin.x < 0.0 && 0.0 < box.pmax.x)
    {
      glVertex2d (0.0, box.pmin.y);
      glVertex2d (0.0, box.pmax.y);
    }

    if (box.pmin.y < 0.0 && 0.0 < box.pmax.y)
    {
      glVertex2d (box.pmin.x, 0.0);
      glVertex2d (box.pmax.x, 0.0);
    }
  }
  glEnd();

  // нарисовать многоугольник
  scene.paint();

  glFinish();
  glutSwapBuffers();
}


/// изменён размер окна
void reshape(int w, int h)
{
  // изменить окно камеры
  camera.forceAspect((double)w / h);
  camera.apply();

  // уведомить OpenGL о новых размерах области рисования
  glViewport (0, 0, w, h);
}


/// обработать нажатие "обычной" клавиши
void keyPressed(unsigned char key, int x, int y)
{
  switch (key)
  {
  case 27: // ESC ASCII code
    glutLeaveMainLoop(); // выйти из цикла обработки событий
    return;

  case '+':
    camera.zoomIn(0.1);
    break;

  case '-':
    camera.zoomOut(1. / 9.);
    break;

  case ' ': case '\r': case '\n':
    {
      // загрузить файл (ввести название в консоль)
      std::string fn;
      std::cout << "Enter the input file name:\n";
      std::getline(std::cin, fn);
      scene.load(fn);

      camera.reset(scene.box());      
    }
    break;
  
  default:
    return; // do nothing for other keys
  }

  camera.apply ();
  glutPostRedisplay ();
}


/// обработать нажатие "специальной" клавиши
void specialKeyPressed(int key, int x, int y)
{
  Vec2 delta;
  switch (key)
  {
  case GLUT_KEY_LEFT:   delta.x = -0.1; break;
  case GLUT_KEY_UP:     delta.y =  0.1; break;
  case GLUT_KEY_RIGHT:  delta.x =  0.1; break;
  case GLUT_KEY_DOWN:   delta.y = -0.1; break;
  case GLUT_KEY_HOME:   camera.reset(scene.box()); break;
  default:
    return; // ничего не делать, если была нажата какая-либо другая клавиша
  }

  delta *= camera.box().extents();
  camera.move(delta);
  camera.apply();
  glutPostRedisplay();  // уведомить GLUT, что следует перерисовать сцену
}


/// по возможности не грузим процессор зря :)
void idle()
{
  avo::sleep(0.05);
}


//
// main

int main(int argc, char **argv)
{
  glutInit (&argc, argv);
  glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  
  glutInitWindowPosition (10, 20);
  glutInitWindowSize (1000, 640);
  glutCreateWindow ("GLUPoly");

  glClearColor (0.0f, 0.0f, 0.07f, 1.0f);

  glEnable (GL_BLEND); // прозрачность
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  glEnable (GL_COLOR_MATERIAL);
  glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

  // включить массивы вершин (OpenGL 1.1)
  glEnableClientState (GL_VERTEX_ARRAY);

  // регистрация колбеков
  glutDisplayFunc  (repaint);
  glutReshapeFunc  (reshape);
  glutKeyboardFunc (keyPressed);
  glutSpecialFunc  (specialKeyPressed);
  glutIdleFunc     (idle);

  // старт
  std::cout << "Starting GLUPoly...\n";
  scene.load("input.txt");
  camera.reset(scene.box());
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
