/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Vec2.hpp
 * @author Dmitri Kuvshinov
 * @brief simple 2D vector implementation within ISO C++ standard
 */
#pragma once
#include <cmath>

namespace avo
{

/// 2D vector
template <class Float>
struct Vec2
{
  using Component = Float;
  using This = Vec2<Float>;
  Float x, y;
  
  explicit Vec2(Float nx = Float(0), Float ny = Float(0))
    : x(nx), y(ny) {}

  explicit Vec2(const Float p[2])
    : x(p[0]), y(p[1]) {}

  This& chs()
  {
    x = -x;
    y = -y;
    return *this;
  }

  This operator-() const
  {
    return This(-x, -y);
  }

  This& operator+=(const This &other)
  {
    x += other.x;
    y += other.y;
    return *this;
  }

  This& operator-=(const This &other)
  {
    x -= other.x;
    y -= other.y;
    return *this;
  }

  /// componentwise multiplication
  This& operator*=(const This &other)
  {
    x *= other.x;
    y *= other.y;
    return *this;
  }

  This& operator*=(const Float coef)
  {
    x *= coef;
    y *= coef;
    return *this;
  }

   
  /// componentwise division
  This& operator/=(const This &other)
  {
    x /= other.x;
    y /= other.y;
    return *this;
  }

  /// possibly slow, use v *= (1.0 / coef) instead for speed
  This& operator/=(const Float coef)
  {
    x /= coef;
    y /= coef;
    return *this;
  }

  /// this dot this, squared length
  Float dotSquare() const
  {
    return x * x + y * y;
  }

  /// vector length
  Float length() const
  {
    return hypot(x, y); // ISO C
  }

  /// normalize this vector in place
  This& normalize()
  {
    const Float len = length();
    if (len != Float(0))
      *this *= Float(1) / len;
    return *this;
  }

  /// store per-coordinate minimum
  This& minimize(const This &other)
  {
    if (other.x < x) x = other.x;
    if (other.y < y) y = other.y;
    return *this;
  }

  /// store per-coordinate maximum
  This& maximize(const This &other)
  {
    if (x < other.x) x = other.x;
    if (y < other.y) y = other.y;
    return *this;
  }

  /// check if this has greater (or equal) coordinates than the other one
  bool dominates(const This &other) const
  {
    return other.x <= x && other.y <= y;
  }
};


template <class Float>
inline bool operator==(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return a.x == b.x && a.y == b.y;
}

template <class Float>
inline bool operator!=(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return !(a == b);
}


/// lexicographic comparison: x then y
struct LexCompareXY
{
  template <class Vec>
  bool operator()(const Vec &a, const Vec &b) const
  {
    return a.x < b.x ||
         (a.x == b.x && a.y < b.y);
  }
};


/// calculate two vectors sum
template <class Float>
inline Vec2<Float> operator+(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return Vec2<Float>(a.x + b.x, a.y + b.y);
}

/// calculate two vectors difference
template <class Float>
inline Vec2<Float> operator-(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return Vec2<Float>(a.x - b.x, a.y - b.y);
}

/// componentwise vector multiplication
template <class Float>
inline Vec2<Float> operator*(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return Vec2<Float>(a.x * b.x, a.y * b.y);
}

/// scale a vector by a coefficient
template <class Float>
inline Vec2<Float> operator*(Float coef, const Vec2<Float> &v)
{
  return Vec2<Float>(coef * v.x, coef * v.y);
}

/// scale a vector by a coefficient
template <class Float>
inline Vec2<Float> operator*(const Vec2<Float> &v, Float coef)
{
  return Vec2<Float>(coef * v.x, coef * v.y);
}

/// componentwise vector division
template <class Float>
inline Vec2<Float> operator/(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return Vec2<Float>(a.x / b.x, a.y / b.y);
}

/// possibly slow, use v * (1.0 / coef) instead for speed
template <class Float>
inline Vec2<Float> operator/(const Vec2<Float> &v, Float coef)
{
  return Vec2<Float>(v.x / coef, v.y / coef);
}

// textual I/O just as number pairs

template <class Float>
inline std::ostream& operator<<(std::ostream &os, const Vec2<Float> &v)
{
  return os << v.x << ' ' << v.y;
}

template <class Float>
inline std::istream& operator>>(std::istream &is, Vec2<Float> &v)
{
  return is >> v.x >> v.y;
}


/// dot product of two vectors
template <class Float>
inline Float dot(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return a.x * b.x + a.y * b.y;
}

/// determinant of [a b] matrix
template <class Float>
inline Float cross(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return a.x * b.y - a.y * b.x;
}

/// compute Euclidian distance
template <class Float>
inline Float distance(const Vec2<Float> &a, const Vec2<Float> &b)
{
  return (a - b).length();
}

/// calculate triangle center (medians intersection point)
template <class Float>
inline Vec2<Float> tricenter(const Vec2<Float> &a, const Vec2<Float> &b, const Vec2<Float> &c)
{
  return Float(1./3.) * (a + b + c);
}

/// compute by-coordinate maximima and minima over an array of Vec2
/**
 * n must be > 0
 */
template <class Float>
inline void bbox2(const Vec2<Float> arr[], std::size_t n, Vec2<Float> &pmin, Vec2<Float> &pmax)
{
  if (n == 0)
    return;

  Vec2<Float> cmax = arr[0], cmin = cmax;
  for (std::size_t i = 1; i < n; ++i)
  {
    cmax.maximize (arr[i]);
    cmin.minimize (arr[i]);
  }

  pmin = cmin;
  pmax = cmax;
}


using Vec2f = Vec2<float>;
using Vec2d = Vec2<double>;

}
