﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Pencil
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 *
 * Реализации игры "Жизнь" Конвея в виде поэлементного сумматора.
 * См. также CellularAutomata.h.
 */
#include <memory>
#include <cstdint>
#include <climits>
#include "CellularAutomata.hpp"


class ConwaysLifeCellwiseSummator 
  : public IBounded2DCellularAutomaton
{
#ifdef _M_X64 // 64-bit x86 CPU target / MSVC
  using Chunk = std::uint64_t;
  static const unsigned CHUNK_BITS = 64;
#else // по умолчанию совпадает с unsigned long
  using Chunk = unsigned long;
  static const unsigned CHUNK_BITS = sizeof(Chunk) * CHAR_BIT;
#endif

  using Data = std::unique_ptr<Chunk[]>; // автоматически управляемый динамический массив бит
  Data cells_old, cells_new;
  // размер >= sizes.first * sizes.second / CHUNK_BITS
  Extents sizes;
  Index aligned_line_width;

public:

  ConwaysLifeCellwiseSummator()
    : sizes(0, 0), aligned_line_width(0), cells_old(nullptr), cells_new(nullptr)
  {}

  /// название автомата
  const char* name() override
  {
    return "Conway's Life: cellwise summator";
  }

  /// получить текущие размеры поля
  Extents extents() const override
  {
    return sizes;
  }

  /// обновить размеры поля
  void reshape(Index width, Index height) override
  {
    if (width > 0 && height > 0)
    {
      sizes.first = width;
      sizes.second = height;
      // выравниваем строчки так, чтобы каждая строчка начиналась с 0-го бита чанка,
      // aligned_line_width делится нацело на CHUNK_BITS
      aligned_line_width = (sizes.first + (CHUNK_BITS - 1)) & (~CHUNK_BITS + 1);
      const auto data_size = (aligned_line_width * sizes.second) / CHUNK_BITS; 
      cells_old.reset(new Chunk[data_size]);
      cells_new.reset(new Chunk[data_size]);
    }
  }

private:

  // извлечь заданный бит из строчки
  static int read_bit(const Chunk *line, const Index x)
  {
    return (line[x / CHUNK_BITS] & (Chunk(1) << (x % CHUNK_BITS))) != 0? 1: 0;
  }

  // записать значение заданного бита в строчке
  static void write_bit(Chunk *line, const Index x, const int value)
  {
    Chunk *const chunk = line + x / CHUNK_BITS;
    const auto shift = x % CHUNK_BITS;
    const auto mask = ~(Chunk(1) << shift);
    const auto bit = Chunk(value) << shift;
    *chunk = (*chunk & mask) | bit;
  }

  // Conway's Life rule
  static int cl_rule(const int previous, const int sum)
  {
    if (previous) // клетка погибает, если имеет < 2 или > 3 соседей
    {
      if (sum < 2 || sum > 3)
        return 0;
    }
    else if (sum == 3)  // место с 3 соседними клетками занимается новой клеткой
    {
      return 1;
    }
    return previous;
  }

  // обработать крайнюю строчку (первую или последнюю)
  void edge_line(const Chunk *src_line, const Chunk *adj_src_line, Chunk *dst_line)
  {
    const Index line_last = sizes.first - 1;

    // левый угол
    write_bit(dst_line, 0, cl_rule(read_bit(src_line, 0),
                                  read_bit(src_line    , 1) + 
      read_bit(adj_src_line, 0) + read_bit(adj_src_line, 1)));

    // основная часть строчки
    for (std::ptrdiff_t i = 1; i < line_last; ++i)
      write_bit(dst_line, i, cl_rule(read_bit(src_line, i),
        read_bit(src_line    , i - 1) +                             read_bit(src_line    , i + 1) +
        read_bit(adj_src_line, i - 1) + read_bit(adj_src_line, i) + read_bit(adj_src_line, i + 1)));

    // правый угол
    write_bit(dst_line, line_last, cl_rule(read_bit(src_line, line_last),
      read_bit(src_line    , line_last - 1) +
      read_bit(adj_src_line, line_last - 1) + read_bit(adj_src_line, line_last)));
  }

  // перевести строчку бит в пиксели
  static void bits_to_pixels(const Chunk *src, const Index width, avo::Pixel *dst)
  {
    for (Index j = 0; j < width; ++j)
      dst[j].binary = read_bit(src, j)? ~avo::PixelInt(0): 0;
  }

public:

  /// построить исходное поле на основе 32-битной картинки (чёрный => 0, нечёрный => 1)
  void fromPixels(const avo::Pixel *src) override
  {
    const Index lines = sizes.second,
                line_width = sizes.first;

    Chunk *const dst = cells_old.get();

    // строчка за строчкой
    #pragma omp parallel for
    for (Index i = 0; i < lines; ++i)
    {
      const auto src_line = src + i * line_width; // источник
      const auto dst_line = dst + i * (aligned_line_width / CHUNK_BITS);
      for (Index j = 0; j < line_width; ++j) // не оптимизировано (было бы эффективнее накапливать один Chunk
        write_bit(dst_line, j, (src_line[j].binary != 0? 1: 0));  // и записывать его в dst_line один раз)
    }
  }

  /// следующий шаг по времени, сформировать 32-битную картинку в dst
  void next(avo::Pixel *dst) override
  {
    const Index lines_1 = sizes.second - 1,
                line_width = sizes.first,
                line_last = line_width - 1;
    
    // первая строчка
    {
      const Chunk *const src_line = cells_old.get();
      edge_line(src_line, src_line + aligned_line_width / CHUNK_BITS, cells_new.get());
      bits_to_pixels(src_line, line_width, dst);
    }

    // последняя строчка
    {
      const Index offset = (aligned_line_width * lines_1) / CHUNK_BITS;
      const Chunk *const src_line = cells_old.get() + offset;
      edge_line(src_line, src_line - aligned_line_width / CHUNK_BITS, cells_new.get() + offset);
      bits_to_pixels(src_line, line_width, dst + line_width * lines_1);
    }

    // все остальные строчки -- единообразно в цикле
    #pragma omp parallel for
    for (Index i = 1; i < lines_1; ++i)
    {
      const Index delta  = aligned_line_width / CHUNK_BITS,
                  offset = delta * i;
                           
      const Chunk *const src_line = cells_old.get() + offset;
      const Chunk *const prev_line = src_line - delta;
      const Chunk *const next_line = src_line + delta;
            Chunk *const dst_line = cells_new.get() + offset;

      // первая клетка
      write_bit(dst_line, 0, cl_rule(read_bit(src_line, 0),
        read_bit(prev_line, 0) + read_bit(prev_line, 1) +
                                 read_bit(src_line , 1) +
        read_bit(next_line, 0) + read_bit(next_line, 1)));

      for (Index j = 1; j < line_last; ++j)
        write_bit(dst_line, j, cl_rule(read_bit(src_line, j),
          read_bit(prev_line, j - 1) + read_bit(prev_line, j) + read_bit(prev_line, j + 1) +
          read_bit(src_line , j - 1) +                          read_bit(src_line , j + 1) +
          read_bit(next_line, j - 1) + read_bit(next_line, j) + read_bit(next_line, j + 1)
        ));

      // последняя клетка
      write_bit(dst_line, line_last, cl_rule(read_bit(src_line, line_last),
        read_bit(prev_line, line_last - 1) + read_bit(prev_line, line_last) +
        read_bit(src_line , line_last - 1) +
        read_bit(next_line, line_last - 1) + read_bit(next_line, line_last)));

      // заполнить строчку пикселей
      bits_to_pixels(src_line, line_width, dst + line_width * i);
    }

    cells_new.swap(cells_old);
  }
};


IBounded2DCellularAutomaton* newConwaysLifeCellwiseSummator()
{
  return new ConwaysLifeCellwiseSummator;
}
