﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Plasma
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Эффект "плазмы". См. http://en.wikipedia.org/wiki/Plasma_effect
 * Демонстрирует использование табличных функций (без интерполяции),
 * шаблонов функций.
 *
 * Нажатие:
 * цифры -- выбор алгоритма;
 * P     -- показывать/скрыть палитру;
 * [     -- уменьшить число потоков OpenMP;
 * ]     -- увеличить число потоков OpenMP;
 * ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <cmath>
#include <iostream>
#include <algorithm>     // max
#include <omp.h>

#include <avo/Pixel.hpp> // RGBA-цвет в 32-битном формате
#include <avo/Timer.hpp>

using avo::Pixel;

//
// константы и глобальные переменные
//

/// размеры картинки: ширина и высота (для простоты зафиксированы)
int width = 800, height = 600;

/// счётчик кадров (сбрасывается каждые 10с)
int frames;
/// момент времени, с которого происходит накопление счётчика кадров
double frames_reported_last;

/// номер текущего кадра -- "время"
int t = 0;

/// признак "показывать палитру"
bool show_palette;

/// сама картинка
Pixel *pixels;

/// период целочисленной синусоиды
const int SinePeriod = 512;
// должен быть степенью двойки
static_assert((SinePeriod & (SinePeriod - 1)) == 0, "Plasma: SinePeriod must be power of 2");


/// палитра, глобальный объект
class Palette
{
public:
  /// конструктор: заполнить палитру
  Palette()
  {
    // линейная модель, аналогично Hue в HSV
    // см. http://upload.wikimedia.org/wikipedia/commons/5/5d/HSV-RGB-comparison.svg?uselang=ru (здесь в обратном порядке)
    //     http://ru.wikipedia.org/wiki/HSV_(цветовая_модель)
    for (int i = 0; i < 64; ++i)
    {
      const uint8_t i4 = uint8_t(i * 4), inv_i4 = uint8_t(255 - i4);
      p[i      ]  = Pixel::rgb( 255   , 0     , i4     );  // красный  -> маджента 
      p[i +  64]  = Pixel::rgb( inv_i4, 0     , 255    );  // маджента -> синий
      p[i + 128]  = Pixel::rgb( 0     , i4    , 255    );  // синий    -> циан
      p[i + 192]  = Pixel::rgb( 0     , 255   , inv_i4 );  // циан     -> зелёный
      p[i + 256]  = Pixel::rgb( i4    , 255   , 0      );  // зелёный  -> жёлтый
      p[i + 320]  = Pixel::rgb( 255   , inv_i4, 0      );  // жёлтый   -> красный
    }
  }

  /// получить цвет по индексу
  Pixel operator[](const int index) const
  { return p[index]; }

private:
  Pixel p[384];
} palette; // конструктор вызывается автоматически


//
// реализации синусоиды
//


/// непосредственно вычисляемая синусоида, отображает [0, SinePeriod-1] -> [0, 255], продолжена периодически на все целые
struct Sine // структура нужна, чтобы можно было просто передать тип шаблону plasma в качестве реализации
{           // "просто" функцию так передать нельзя
  int operator()(const int x) const
  {
    static const float PI2 = 6.283185307179586476925286766559f / SinePeriod;
    const float nx = (x & (SinePeriod - 1)) * PI2;
    return (int)(127.5f * (std::sin(nx) + 1.f));
  }
};

/// табличная синусоида, отображает [0, SinePeriod-1] -> [0, 255], продолжена периодически на все целые
class TableSine
{
public:
  static const int N = SinePeriod;
  TableSine()
  {
    Sine sine; // повторное использование ;)
    for (int i = 0; i < N; ++i)
      p[i] = sine(i);
  }

  int operator()(const int x) const
  { return p[x & (N - 1)]; }

private:
  int p[N];
};


//
// реализации функции расстояния
//

struct SqrtDist
{
  int operator()(int a, int b) const
  {
    return (int)std::sqrt(float(a * a + b * b));
  }
};

struct CrudeDist
{
  int operator()(int a, int b) const
  {
    // грубая, но очень быстро работающая аппроксимация:
    // вместо окружностей получаем восьмиугольники
    a = std::abs(a);
    b = std::abs(b);
    return a < b? b + (a >> 1): a + (b >> 1);
  }
};



//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{        /* ширина,  высота, тип пикселя, тип компонента, указатель на картинку */
  glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);

  // счётчик кадров в секунду, выводит среднее за 10с
  ++frames;
  const double dt = avo::seconds() - frames_reported_last;
  if (dt >= 10.0)
  {
    std::cout << frames / dt << '\n';
    frames = 0;
    frames_reported_last += dt;
  }

  glFinish(); // дождаться завершения обработки очереди команд OpenGL
}


/// изменение размеров окна, функция получает новые размеры
void reshape(int w, int h)
{
  // заменим старое содержимое pixels на новое
  delete[] pixels;
  pixels = new Pixel[(w < 384? 384: w) * h];

  // установим новые размеры картинки
  width   = w;
  height  = h;
    
  // установить новые координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}


/// рисование плазмы, использующее переданные функторы-реализации синуса и расстояния
template <class SineImpl, class DistImpl>
void plasma()
{
  static const SineImpl sine;
  static const DistImpl dist;

  t++;
  // все константы взяты "с потолка"
  const int mwh = std::max(width, height) / 2,
            st1 = (mwh * sine(    t     +  17)) / 256,
            st2 = (mwh * sine(3 * t / 2 + 107)) / 256,
            st3 = (mwh * sine(3 * t / 8 + 210)) / 256,

  x0 = (st1 -     width / 7) * 3,   y0 = (2 * height / 3 - st3) * 2,
  x1 = (st2 + 3 * width / 4)    ,   y1 = (4 * height / 3 - st1) * 7,
  x2 = (6 * width / 7 - st3) * 2,   y2 = (    height / 2 - st1),
  x3 = (3 * width / 5 - st1)    ,   y3 = (st2 - st3) / 2;

  #pragma omp parallel for
  for (int y = 0; y < height; ++y)
  {
    Pixel *line = pixels + y * width;

    if (show_palette)
    {
      for (int x = 0; x < 384; ++x)
        line[x] = palette[x];
    }

    const int z0 = y - y0,  z1 = y - y1,
              z2 = y - y2,  z3 = y - y3;

    for (int x = show_palette? 384: 0; x < width; ++x)
    {	
      const int v0 = x - x0,  v1 = x - x1,
                v2 = x - x2,  v3 = x - x3;

      const int d0 = dist(v0, z0),
                d1 = dist(v1, z1),
                d2 = dist(v2, z2),
                d3 = dist(v3, z3);

      const int tv  = sine(d0) + sine(d1) + sine(d2) + sine(d3);
      line[x] = palette[(384 * tv) / 1024];
    }
  }

  glutPostRedisplay();
}

// TASK: вместо эффекта "плазмы" реализовать другой эффект, например, "бегущие волны".


/// вывести в консоль количество потоков в рабочей группе OpenMP
void report_current_threads()
{
  std::cout << "\nThreads: " << omp_get_max_threads() << std::endl;
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void handle_key(unsigned char key, int x, int y)
{
  switch(key) // код клавиши
  {
  case '1':
    std::cout << "\ndirect implementation\n";
    glutIdleFunc(plasma<Sine, SqrtDist>);
    break;

  case '2':
    std::cout << "\ntable implementation\n";
    glutIdleFunc(plasma<TableSine, SqrtDist>);
    break;

  case '3':
    std::cout << "\ntable/crude dist implementation\n";
    glutIdleFunc(plasma<TableSine, CrudeDist>);
    break;

  case 'p': case 'P':
    show_palette = !show_palette;
    break;

  case '[': // убрать поток из рабочей группы OpenMP
    {
      const int threads = omp_get_max_threads();
      if (threads > 1)
        omp_set_num_threads(threads - 1);
      report_current_threads();
    }
    break;

  case ']': // добавить поток в рабочую группу OpenMP
    {
      const int threads = omp_get_max_threads();
      if (threads < omp_get_num_procs())
        omp_set_num_threads(threads + 1);
      report_current_threads();
    }
    break;

  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  default:
    return;
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки

  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_SINGLE); // режим: 32-битный цвет RGBA, одинарный буфер кадра SINGLE
  glutInitWindowPosition(0, 0);      // координаты верхнего левого угла окна на экране
  glutInitWindowSize(width, height); // полные ширина и высота окна
  glutCreateWindow("Plasma");        // создать окно, в функцию передаётся заголовок создаваемого окна

  // привязка функций-обработчиков событий
  glutIdleFunc(plasma<TableSine, SqrtDist>);
  glutDisplayFunc(repaint);     // вызывать для отрисовки кадра
  glutKeyboardFunc(handle_key); // вызывать для обработки нажатий "обычных" клавиш
  glutReshapeFunc(reshape);     // изменение размеров окна
  
  // старт
  pixels = new Pixel[width * height];
  frames_reported_last = avo::seconds();
  std::cout << "Starting Plasma\n";
  report_current_threads();
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
