﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: SortVis
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Визуализатор различных алгоритмов сортировки.
 * Также см. https://evetro.wordpress.com/2011/04/28/av-1/
 * Нажатие:
 * +/-   -- ускорение/замедление демонстрации;
 * ENTER -- ввод параметров из консоли:
 *   скорость (количество событий в секунду),
 *   количество,
 *   набор алгоритмов, заканчивающийся точкой с запятой (просто ; для повтора);
 * 1, 2, 3 ... -- выбор типа последовательности;
 * SPACE -- перезапуск процесса сортировки;
 * P     -- пауза;
 * T     -- включить/выключить отображение названий алгоритмов;
 * ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <cstdlib>     // rand/srand
#include <string>
#include <iostream>
#include <iomanip>
#include <vector>      // динамический массив
#include <algorithm>   // std::sort + некоторые вспомогательные элементы
#include <functional>  // std::mem_fn, std::function
#include <numeric>     // std::accumulate
#include <iterator>    // std::iterator_traits
#include <map>         // словарь алгоритмов сортировки и очередь действий
#include <utility>     // std::move

#include <avo/Color.hpp>
#include <avo/Timer.hpp>
#include "Sorts.hpp"     // реализации алгоритмов сортировки
#include "Sequences.hpp" // реализации генераторов последовательностей

using avo::Color;
using avo::RED;
using avo::ORANGE;
using avo::GREEN;
using avo::GREY;


//
// вспомогательные определения
//

/// отложенное действие, всё, что с ним можно сделать -- "выполнить", применив ()
using Action  = std::function<void()>;
/// контейнер отложенных действий
using Actions = std::vector<Action>;

/// очередь событий ("действий"), упорядоченных по временной метке
/** ActionQueue ничего не знает о природе хранимых в ней действий */
class ActionQueue
{
public:
  /// добавить новый объект действия с заданной временной меткой
  void push(double time, Action action)
  {
    repr.emplace(time, action);
  }

  /// извлечь все отложенные действия с временными метками не больше переданной (перемещаются в actions)
  /** возвращает истину, если какие-либо действия были выполнены */
  bool next(double time, Actions &actions)
  {
    auto p = repr.begin();
    const auto end = repr.upper_bound(time);
    if (p == end)
      return false;
    for (; p != end; ++p)
      actions.emplace_back(std::move(p->second));
    repr.erase(repr.begin(), end);
    return true;
  }

private:
  using Repr = std::multimap<double, Action>; // representation == представление
  Repr repr;
};

/// "загнать" значение в заданный диапазон
template <class Number>
inline Number clamp(Number x, Number lower_bound, Number upper_bound)
{
  if (x < lower_bound) return lower_bound;
  if (upper_bound < x) return upper_bound;
  return x;
}


/// набор строк
using Strings = std::vector<std::string>;


//
// сцена
//

/// прямоугольник со сторонами, параллельными осям координат, который можно нарисовать
struct Rect // визуализирует элемент сортируемой последовательности
{
  float x0, y0; // координаты нижнего левого угла
  float x1, y1; // координаты верхнего правого угла
  Color color;  // цвет заливки прямоугольника

  /// базовый конструктор
  explicit Rect(float x0 = 0.f, float y0 = 0.f,
                float x1 = 0.f, float y1 = 0.f, 
                const Color &col = GREY)
      : x0(x0), y0(y0), x1(x1), y1(y1), color(col)
  {}

  /// OpenGL1-код, рисующий данный прямоугольник
  void paint() const
  {
    color.select();
    glRectf( x0, y0, x1, y1 );
  }
};


/// массив прямоугольников, который можно нарисовать
class RectArray
{
  using Rects = std::vector<Rect>;
  Rects rects;  // набор прямоугольников
  float dx, dy; // сдвиг массива в плоскости

public:
  RectArray() {}

  /// сгенерировать набор прямоугольников, длины которых равны числам из диапазона [begin, end)
  template <class FwdIt>
  RectArray(FwdIt begin, FwdIt end, float dx = 0.0f, float dy = 0.0f)
    : dx(dx), dy(dy)
  {
    // заполнить rects
    while (begin != end)
      rects.emplace_back(0.f, 0.f, *begin++);

    // сместить прямоугольники
    const unsigned sz = rects.size();
    const float rect_height = 0.75f / sz; // высота прямоугольника == три четверти шага по вертикали
    for (unsigned i = 0; i < sz; ++i)
    {
      Rect &r = rects[i];
      r.y0 = r.y1 = float(i);
      r.y1 += 0.75f;
    }
  }

  /// указать цвет для всех прямоугольников
  void clearColor(const Color &col = GREY)
  {
    for (Rect &r: rects)
      r.color = col;
  }

  /// нарисовать все прямоугольники
  void paint() const
  {
    glLoadIdentity();
    glTranslatef( dx, dy, 0.0f );
    for (const Rect &r: rects)
      r.paint();
  }


  // реакция на события

  /// событие "присваивание": [index] = value
  void assign(unsigned index, float value)
  {
    Rect &r = rects.at(index);
    r.color = RED;
    r.x1 = value;
  }

  /// событие "сравнение": [i1] < [i2]?
  void compare(unsigned i1, unsigned i2)
  {
    rects.at(i1).color = GREEN;
    rects.at(i2).color = GREEN;
  }

  /// событие "обмен": [i1] <-> [i2]
  void swap(unsigned i1, unsigned i2)
  {
    rects.at(i1).color = ORANGE;
    rects.at(i2).color = ORANGE;
    std::swap( rects[i1].x1, rects[i2].x1 );
  }
};


/// TimeMapper формирует временные метки событий и служит связующим звеном между набором зондов, очередью событий
class TimeMapper
{
private:
  RectArray   *ra; // связанный массив прямоугольников
  ActionQueue *eq; // связанная очередь событий
  double t, dt;    // начальный момент времени и шаг

  // статистика
  std::size_t stat_assign, stat_comp, stat_swap;

public:
  // для создания объекта требуется передать объект массива прямоугольников и объект очереди событий
  TimeMapper(RectArray &R, ActionQueue &Q, double time = 0., double step = 1.)
    : ra(&R), eq(&Q), t(time), dt(step), stat_assign(0), stat_comp(0), stat_swap(0)
  {}

  /// добавить действие и вычислить следующую временную метку
  void pushNext(Action action)
  {
    eq->push(t, action);
    t += dt;
  }

  // TimeMapper "знает", какие виды событий могут быть,
  // это не очень приятно, но альтернатива -- хранить указать ra прямо в зондах -- мне не очень по душе.

  void pushAssignment(unsigned i, float v)
  {
    ++stat_assign;
    pushNext(std::bind(&RectArray::assign, ra, i, v));
  }

  void pushComparison(unsigned i1, unsigned i2)
  {
    ++stat_comp;
    pushNext(std::bind(&RectArray::compare, ra, i1, i2));
  }

  void pushSwap(unsigned i1, unsigned i2)
  {
    ++stat_swap;
    pushNext(std::bind(&RectArray::swap, ra, i1, i2));
  }

  // извлечение статистики
  std::size_t assignments() const { return stat_assign; }
  std::size_t comparisons() const { return stat_comp; }
  std::size_t swaps() const { return stat_swap; }
};


/// зонд генерирует события в процессе работы алгоритмов сортировки
/**
 * Таким образом, используя "зонды" вместо обычных объектов, можно записывать
 * проделываемые на них операции, чтобы потом визуализировать их.
 */
class Probe
{
  // объект TimeMapper служит связующим звеном между набором зондов, очередью событий
  TimeMapper *tm; // и массивом прямоугольников
  unsigned index; // позиция зонда в массиве
  float value;    // числовое значение зонда (длина прямоугольника)

public:
  /// базовый конструктор
  explicit Probe(unsigned i = 0, float v = 0.f, TimeMapper *tm = nullptr)
    : tm(tm), index(i), value(v)
  {}

  /// сгенерировать событие "присваивание"
  Probe& operator=(const Probe &right)
  {
    if (tm) tm->pushAssignment(index, right.value);
    value = right.value;
    return *this;
  }

  /// сгенерировать событие "сравнение"
  bool operator<(const Probe &right) const
  {
    if (tm) tm->pushComparison(index, right.index);
    return value < right.value;
  }

  // Если требуется сравнивать на равенство,
  // добавить сюда operator==, аналогичный operator<, определённому выше
  // -- тоже событие "сравнение".

  /// сгенерировать событие "обмен"
  void swap(Probe &other)
  {
    if (tm) tm->pushSwap(index, other.index);
    std::swap(value, other.value);
  }
};

/// выбирается вместо базовой std::swap благодаря ADL
inline void swap(Probe &a, Probe &b) 
{
  a.swap(b);
}


/// сцена представляет собой фасад всего "безобразия", описанного выше :)
class Scene
{
private:
  using Probes = std::vector<Probe>;
  using PIt    = Probes::iterator;
  // параллельные массивы:
  std::vector<ActionQueue> queues;     // очереди событий
  std::vector<RectArray>   arrays;     // сортируемые массивы прямоугольников, отвечающие за визуализацию
  Strings                  algorithms; // названия алгоритмов сортировки

  // выбор генератора последовательности
  using Samples = std::vector<float>;
  using SIt     = Samples::iterator;
  using GeneratorFunction = std::function<void(SIt, SIt)>;
  GeneratorFunction generator;

public:
  Scene()
    : height(1.f),
      generator(generateRandom<SIt>),
      is_paused(false),
      show_titles(true),
      time_offset(0.0)
  {
    fillSorts(); // сформировать базовый набор алгоритмов сортировки,
    // чтобы можно было выбирать алгоритм по названию
  }

  /// сбрасывает внутренние структуры, заполняет массивы случайной последовательностью, сортирует
  /**
   * Принимает список названий алгоритмов (строки), заданный диапазоном [algorithm_begin, algorithm_end).
   * Названия см. в fillSorts.
   */
  template <class It>
  void makeNew(unsigned amount, It algorithm_begin, It algorithm_end)
  {
    height = float(amount); // высота одного столбца (используется при рисовании)

    Strings old_algorithms;
    algorithms.swap(old_algorithms); 
    // теперь algorithms пуст, а старый набор алгоритмов сохранён в old_algorithms,
    // оставить только те алгоритмы, которые есть в словаре:
    std::copy_if (algorithm_begin, algorithm_end,
      std::back_inserter(algorithms),
      [this](const std::string &name) { return sorts.count(name) != 0; });
    // если нет годных алгоритмов -- вернуть старый набор:
    if (algorithms.empty())
      algorithms.swap(old_algorithms);
    // если набор алгоритмов оказался пуст -- использовать набор алгоритмов по умолчанию:
    if (algorithms.empty())
      defaultAlgorithms();

    arrays.clear();
    queues.clear();
    // инициализировать массивы прямоугольников и очереди событий
    const unsigned  N = algorithms.size();
    arrays.reserve (N);
    queues.resize  (N);

    Samples samples (amount); // набор псевдослучайных "семплов"

    std::cout << "generating samples ... " << std::flush;
    generator(samples.begin(), samples.end());
    std::cout << "done\n" << std::endl;

    // инициализировать массивы и очереди
    for (unsigned i = 0; i < N; ++i)
    {
      // создать новый массив
      arrays.emplace_back(samples.begin(), samples.end(), float(i));

      // выполнить сортировку
      TimeMapper tm(arrays[i], queues[i]);  // связующее звено
      Probes probes; // сортируемые "зонды"
      probes.reserve(amount);
      for (unsigned j = 0; j < amount; ++j)
        probes.emplace_back(j, samples[j], &tm);

      // собственно сортировка
      std::cout << std::left << std::setw(10) << algorithms[i] << " ... " << std::flush;
      sorts[algorithms[i]] (probes.begin(), probes.end()); // вызов функции из массива
      std::cout << "done: " << std::right;

      // вывести статистику
      std::cout << std::setw(10) << tm.comparisons() << "C, "
                << std::setw(10) << tm.assignments() << "A, "
                << std::setw(10) << tm.swaps() << "S"
                << std::setw(10) <<
      // вес операций: сравнение = 1, чтение = 3, запись = 6
      // comparison => 2 чтения + сравнение = 7,
      // assignment => 0.5("вероятность")*чтение + запись = 7.5,
      // swap => 0.5("вероятность")*два чтения + две записи = 9.
                std::ceil( 7.0 * tm.comparisons() + 7.5 * tm.assignments() + 9.0 * tm.swaps() )
                << "ops" << std::endl;
    }

    std::cout.put('\n');
  }

  /// распечатать набор встроенных алгоритмов сортировки
  void print_supported_sorts() const
  {
    for (auto &s : sorts)
      std::cout << s.first << '\n';
    std::cout << std::endl;
  }

  /// изобразить все очереди
  void paint() const
  {
    // установить матрицу проекции, область рисования -- прямоугольник размера N * 1
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(-0.05, algorithms.size() + 0.05, -0.125, height + 0.875);
    glMatrixMode(GL_MODELVIEW);

    // нарисовать каждый набор
    for (auto &a : arrays)
      a.paint();

    // надписи
    if (show_titles)
    {
      avo::WHITE.select(); // выбрать цвет текста
      for (std::size_t i = 0, isize = algorithms.size(); i != isize; ++i)
      {
        // установить позицию левого нижнего угла отображаемой строки
        glLoadIdentity();
        glRasterPos2f(GLfloat(i), height * 0.09f * (3 + i % 8));
        // изобразить строку текста, используя встроенный растровый шрифт
        glutBitmapString(GLUT_BITMAP_HELVETICA_18, 
          reinterpret_cast<const unsigned char*>(algorithms[i].c_str()));
      }
    }
  }

  /// применить следующее событие (достигшее времени time) каждой очереди, вернуть false, если исчерпаны все очереди
  bool next(double time)
  {
    if (!is_paused)
    {
      time -= time_offset;
      last_time = time;

      // извлечь отложенные действия, сбросить цвета наборов, для которых произошли события
      // queues и arrays "параллельны"
      auto q = queues.begin(), qend = queues.end();
      auto a = arrays.begin(), aend = arrays.end();
      for (; q != qend; ++q, ++a)
      {
        if (q->next(time, deferred))
          a->clearColor();
      }

      // выполнить действия
      for (auto &d : deferred)
        d();

      if (deferred.empty())
        return false;

      deferred.clear();
      return true;
    }

    // на паузе
    time_offset = time - last_time;
    return false; // изменений нет -- перерисовывать нечего
  }

  /// выбрать генератор из предопределённого набора по номеру, сюда добавляются новые генераторы последовательностей
  void anotherGenerator(int n)
  {
    static const GeneratorFunction generators[] =
    {
      generateRandom            <SIt>, 
      generateSorted            <SIt>, 
      generateAntisorted        <SIt>,
      generateBitonic           <SIt>,
      generateAntibitonic       <SIt>,
      generateSortedBlocks      <SIt>,
      generateAntisortedBlocks  <SIt>
    };

    if (0 <= n && n < sizeof(generators)/sizeof(GeneratorFunction))
      generator = generators[n];
  }

  // пауза -- снять с паузы
  void toggle_pause()
  { is_paused = !is_paused; }
  // включить/выключить отображение названий алгоритмов
  void toggle_titles()
  { show_titles = !show_titles; }

private:
  // средства обеспечения паузы поверх текущей реализации
  bool is_paused;
  double last_time, time_offset;
  // ---

  bool show_titles; // показывать названия алгоритмов
  float height; // рабочая высота сцены == количеству сортируемых элементов

  using SortFunction = std::function<void(PIt, PIt)>;
  std::map<std::string, SortFunction> sorts; // позволяет выбрать сортировку по названию во время исполнения

  /// сформировать словарь алгоритмов сортировки (sorts), сюда добавляются новые алгоритмы сортировки
  void fillSorts()
  {
    #ifdef S
    #error Macro redefinition, check the code.
    #endif
    #define S(name, algo) \
      sorts[name] = algo<PIt>
    S("bubble",     bubbleSort);
    S("cocktail",   cocktailSort);
    S("comb",       combSort);
    S("minmax",     minmaxSort);
    S("insert",     insertionSort);
    S("quick",      quickSort);
    S("merge",      mergeSort);
    S("merge2",     mergeSort2);
    S("merge3",     mergeSort3);
    S("merge4",     mergeSort4);
    S("rbitonic",   bitonicSort);
    S("bitonic",    bitonicSort2);
    S("oddeven",    oddevenSort);
    S("pairwise",   pairwiseSort);
    S("heap",       heapSort);
    S("random",     randomTrySort);
    S("ci",         ciSort);
    S("combinsert", ciSort);
    S("cycle",      cycleSort);
    S("bst",        treeSort);
    S("tree",       treeSort);
    S("bbt",        treeSortBalanced);
    S("balanced",   treeSortBalanced);
    S("stdheap",    stdHeapSort);
    S("hybheap",    hybridHeapSort);
    S("std",        std::sort);
    S("stable",     std::stable_sort);
    #undef S
  }

  /// набор алгоритмов сортировки "по умолчанию"
  void defaultAlgorithms()
  {
    const char* const ss[] =
    { "comb", "quick", "stable", "rbitonic", "merge", "heap", "insert" };
    algorithms.assign (ss, ss + sizeof(ss)/sizeof(char*));
  }

  // накапливаемые отложенные действия
  Actions deferred;
};


/// "скорость" таймера, хранит множитель
class Rate
{
  double factor;

public:
  explicit Rate(double f = 1.0)
    : factor(f) {}

  void multiply(double f)
  { factor *= f; }

  double operator()(double dt) const
  { return dt * factor; }
};


//
// глобальные переменные
//

/// количество сортируемых элементов
unsigned elements_to_be_sorted = 64;

/// используемые алгоритмы сортировки
Strings algorithms;

/// таймер очередей
avo::Timer<Rate> timer;

/// объект сцены
Scene scene;


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{
  glClear(GL_COLOR_BUFFER_BIT);
  scene.paint();
  glFinish();
  glutSwapBuffers();
}


/// вычислить состояние сцены на следующий кадр
void frameMove()
{
  if (scene.next(timer.mark())) // если удалось выбрать следующее событие
    glutPostRedisplay();  // необходимо перерисовать
}


/// перезапуск (сортировать заново)
void restart()
{
  scene.makeNew( elements_to_be_sorted, algorithms.begin(), algorithms.end() );
  timer.restart();
}

/// обработать нажатие "обычной" клавиши
void keyPressed(unsigned char key, int x, int y)
{
  switch (key)
  {
  case '+': // ускорить демонстрацию в 1.5 раза
    timer.rate_proxy().multiply (1.5);
    break;

  case '-': // замедлить демонстрацию в 1.5 раза
    timer.rate_proxy().multiply (2./3.);
    break;

  case 27: case 'q': case 'Q': case 'й': case 'Й':
    glutLeaveMainLoop();
    break;

  case ' ':
    restart();
    break;

  case 'p': case 'P': case 'з': case 'З':
    scene.toggle_pause();
    break;

  case 't': case 'T': case 'е': case 'Е':
    scene.toggle_titles();
    glutPostRedisplay();
    break;

  // выбор генератора + рестарт
  case '1': case '2': case '3': case '4': case '5':
  case '6': case '7': case '8': case '9':
    scene.anotherGenerator(key - '1');
    restart();
    break;

  case '0':
    scene.anotherGenerator(10);
    restart();
    break;

  case '\r': case '\n':
    // прочитать:
    {
      using std::cin;
      using std::cout;
      // ... скорость,
      double speed = 20.;
      cout << "\nspeed = ";
      cin >> speed;
      if (cin)
      {
        speed = clamp(speed, 0.1, 1e+6);
        cout << " => " << speed << '\n';
        timer.rate_proxy() = Rate(speed);
      }
      else
      {
        cin.clear();
        cin.ignore();
      }

      // ... количество,
      int elms = 64;
      cout << "\nelements = ";
      cin >> elms;
      if (cin)
      {
        elements_to_be_sorted = clamp(elms, 4, 8192);
        cout << " => " << elements_to_be_sorted  << '\n';
      }
      else
      {
        cin.clear();
        cin.ignore();
      }

      // ... список алгоритмов сортировок
      cout << "\nAlgorithms (delimit with newline, finish with semicolon):\n";
      algorithms.clear();
      while (true)
      {
        std::string name;
        std::getline (cin, name);
        if (!cin.good() || name == "end" || name == ";")
          break;
        algorithms.emplace_back (std::move(name));
      }
      cout << std::endl;
    }

    restart();
    break;

  default:
    return;
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки
  std::srand ((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(800, 593);
  glutCreateWindow("Sort visualization");

  // параметры OpenGL для данного примера
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  
  // привязка функций-обработчиков событий
  glutDisplayFunc(repaint);
  glutKeyboardFunc(keyPressed);
  glutIdleFunc(frameMove);

  // старт
  std::cout << "Starting SortVis...\n";
  timer.rate_proxy().multiply (20.); // 20 событий в секунду
  scene.print_supported_sorts();
  restart();
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
