﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Game.hpp
 *
 * Модель игры. Реализация движения змейки.
 */
#pragma once
#include <algorithm>
#include <vector>
#include <utility>
#include <limits>

#include "ViewModelBase.hpp"


/// модель игрового поля, хранит массив ячеек Cell
/**
 * Массив ячеек организован в "матрицу", которую можно адресовать
 * индексами (x, y) с помощью оператора ().
 */
class Field
{
public:
  /// координаты на поле
  /** 
   * Работа с ними опирается на "оборачивающую" арифметику в дополнительном коде
   * (переполнения и преобразование отрицательных к беззнаковому типу).
   */
  using Position = std::pair<Index, Index>;

  /// задать размеры поля по горизонтали и вертикали
  /**
   * Заполняет поле пустыми ячейками.
   *
   * @param horz горизонтальный размер, индекс x от 0 до horz - 1
   * @param vert вертикальный размер, индекс y от 0 до vert - 1
   */
  Field(Index horz, Index vert)
    : xsize_(horz), ysize_(vert), cells(horz * vert, Cell_Empty)
  {}

  /// получить размер по горизонтали
  Index xsize() const
  {
    return xsize_;
  }

  /// получить размер по вертикали
  Index ysize() const
  {
    return ysize_;
  }

  /// получить клетку по координатам, бросает исключение для плохих x, y
  Cell operator()(Index x, Index y) const
  {
    return cells.at(x + xsize_ * y);
  }
  
  /// получить клетку по координатам, бросает исключение для плохих x, y
  Cell& operator()(Index x, Index y)
  {
    return cells.at(x + xsize_ * y);
  }

  /// получить клетку по координатам, бросает исключение для плохих координат
  Cell operator[](const Position &p) const
  {
    return (*this)(p.first, p.second);
  }

  /// получить клетку по координатам, бросает исключение для плохих координат
  Cell& operator[](const Position &p)
  {
    return (*this)(p.first, p.second);
  }
  

  /// пройтись по всем ячейкам поля, visitor принимает параметры (dx, dy, cell)
  /**
   * В начале перебора: dx == 0 и dy == 0;
   * затем передаются смещения в клетках от предыдущей переданной клетки.
   * Так сделано для упрощения использования функции glRasterPos.
   */
  template <class Visitor>
  void scan(Visitor visitor) const
  {
    visitor (0, 0, cells[0]);
    for (Index x = 0, i = 1, sz = cells.size(); i < sz; ++i)
    {
      if (++x == xsize_)
      {
        visitor(1 - xsize_, 1, cells[i]);
        x = 0;
      }
      else
      {
        visitor(1, 0, cells[i]);
      }
    }
  }

private:
  Index xsize_, ysize_;    // размеры поля
  std::vector<Cell> cells; // ячейки поля
};


/// состояние "змеи"
class Snake
{
public:
  /// позиция на поле
  using Position = Field::Position;

  /// по умолчанию двигается из левого нижнего угла (0, 0) вверх по y
  /**
   * @param field поле, на котором разместить змею
   * @param snake_len длина змеи, змея укладывается горизонтально
   */
  explicit Snake(Field &field, Index snake_len = 5)
    : field(field)
  {
    createSnake(snake_len);
  }

  /// сбросить змею
  /**
   * Змея удаляется с поля field, затем действует как конструктор.
   */
  void reset(Index snake_len = 0)
  {
    for (const auto &p : positions)
      field[p] = Cell_Empty;
    createSnake(snake_len);
  }

  /// двигаться "вверх"
  void up()
  {
    vx =  0;
    vy = +1;
    check_reverse();
  }

  /// двигаться "вниз"
  void down()
  {
    vx =  0;
    vy = -1;
    check_reverse();
  }
  
  /// двигаться "влево"
  void left()
  {
    vx = -1;
    vy =  0;
    check_reverse();
  }

  /// двигаться "вправо"
  void right()
  {
    vx = +1;
    vy =  0;
    check_reverse();
  }

  /// передвинуться на один шаг вперёд
  /**
   * @return пройденная клетка (если пустая -- прошла, если змея -- столкновение, кролик -- съел)
   */
  Cell step()
  {
    const Position new_head = moved(positions.front());
    if (field.xsize() <= new_head.first || field.ysize() <= new_head.second)
      return Cell_Snake;
    return warpHead(new_head);
  }

  /// телепортировать голову змеи
  /**
   * @return старое значение клетки, которую заняла голова змеи
   */
  Cell warpHead(const Position &new_head)
  {
    switch (const Cell next_cell = field[new_head])
    {
    case Cell_Snake:
      return Cell_Snake;

    case Cell_Empty:
      field[positions.back()] = Cell_Empty;
      positions.pop_back ();
      // fall-through
    default:
      field[new_head] = Cell_Snake;
      positions.insert (positions.begin(), new_head);
      return next_cell;
    }
  }

  /// get snake length
  Index length() const { return positions.size(); }

private:

  using Positions = std::vector<Position>;
  /// позиции клеток, занимаемых змеёй, голова на индексе 0
  Positions positions;
  Index vx, vy; // скорость по x, по y
  Field &field; // привязка к полю

  /// создать змею
  void createSnake(Index snake_len)
  {
    vx = 0;
    vy = 1;
    positions.resize(snake_len);
    for (Index i = 0; i < snake_len; ++i)
    {
      const Position block(snake_len - i - 1, 0);
      positions[i] = block;
      field[block] = Cell_Snake;
    }
  }

  /// сдвинутая на vx, vy позиция
  Position moved(const Position &p) const
  {
    return Position(p.first + vx, p.second + vy);
  }

  /// обратить порядок, если нужно
  void check_reverse()
  {
    if (1 < positions.size() && moved(positions.front()) == positions[1])
    {
      std::reverse(positions.begin(), positions.end());
      // даже после обращения порядка, "голова" может двигаться в "шею"
      if (moved(positions.front()) == positions[1])
        vx = positions.front().first  - positions[1].first ,
        vy = positions.front().second - positions[1].second;
    }
  }
};


/// описание уровня
struct Level
{
  Index start_length, ///< начальная длина змеи
        max_length;   ///< максимальная длина змеи, число кроликов == max_length - start_length
  double speed;       ///< скорость в ячейках в секунду
  Score max_movement, ///< максимальное количество ячеек, засчитываемых в очки
        penalty;      ///< штраф за столкновение (домножается на номер уровня)

  Level()
    : start_length(4),
      max_length(10),
      speed(4.0),
      max_movement(1200),
      penalty(300)
  {}
};


/// увеличиваемое число с автоматически обновляемым текстовым представлением
/**
 * Можно считать данный класс излишеством.
 * Он создан для того, чтобы не вызывать std::to_string на каждом кадре.
 */
class NumberPresenter
{
  // этот класс введён потому, что мне претит на каждом кадре преобразовывать
  // число в текст, пусть это и не окажет сколько-нибудь заметного влияния на скорость
public:
  virtual ~NumberPresenter() {} // actually not needed here
  /*implicit*/ NumberPresenter(Score n = 0);

  /// получить числовое значение
  operator Score() const { return number;}
  /// получить текстовое представление
  const std::string& str() const { return repr; }

  NumberPresenter& operator+=(Score inc);
  NumberPresenter& operator++() { return (*this) += 1; }

protected:
  /// обновить значение и текстовое представление
  virtual void update(Score n);

private:
  Score number;
  std::string repr;
};


/// увеличиваемое и уменьшаемое число с автоматически обновляемым текстовым представлением
/**
 * Дополняет NumberPresenter возможностью уменьшать хранимое число (очки).
 * Также сохраняет достигнутый на текущий момент максимум.
 */
class NumberPresenterDecrementable : public NumberPresenter
{
public:
  NumberPresenterDecrementable(Score n = 0)
    : NumberPresenter(n), max_val(n) {}

  NumberPresenterDecrementable& operator-=(Score inc);
  NumberPresenterDecrementable& operator--() { return (*this) -= 1; }
  /// получить максимальное достигнутое на момент значение
  Score getMax() const { return max_val; }

protected:
  void update(Score n) override;

private:
  Score max_val;
};


/// класс представляющий модель игры "Змейка"
class ZmeikaGame
{
public:
  ZmeikaGame();

  /// временнАя метка последнего кадра (секунды)
  double last_time;

  /// собственно игровое поле
  Field field;

  /// текущий уровень
  NumberPresenter level;

  /// текущие очки игрока
  NumberPresenterDecrementable score;

  /// змея будет двигаться вверх
  void up()    { snake.up();    }
  /// змея будет двигаться вниз
  void down()  { snake.down();  }
  /// змея будет двигаться влево
  void left()  { snake.left();  }
  /// змея будет двигаться вправо
  void right() { snake.right(); }

  /// выполнить следующий шаг, если с предыдущего прошло достаточно времени
  /**
   * Время введено в класс ZmeikaGame, поскольку оно является элементом описания уровня.
   * Может быть, следовало бы полностью делегировать учёт времени наследнику.
   *
   * @param t текущее время в секундах
   * @return true в случае столкновения
   */
  bool step(double t);

  /// начать новую игру
  void start()
  {
    readLevels();
    newRabbit(Cell_Rabbit);
    selectLevel(1);
    score = 0;
  }

  /// телепортировать голову змеи
  /**
   * @return true если произошло столкновение (останов змеи)
   */
  bool warpSnake(const Field::Position &new_head)
  {
    return movedInto(snake.warpHead(new_head));
  }

private:

  /// прочитать описания уровней
  void readLevels();

  /// поставить нового кролика
  void newRabbit(Cell rabbitType);
  /// выбрать уровень
  void selectLevel(Index next_level);

  /// реакция змеи на движение в клетку с заданным содержимым
  /**
   * @return true если произошло столкновение (останов змеи)
   */
  bool movedInto(const Cell);

  /// премиальные за уровень
  Score level_score;
  /// очки за движение в ту или иную клетку (по её коду Cell)
  Score move_score[Cell_Types];
  /// текущая скорость змеи определяется задержкой между кадрами (секунды)
  double time_step;
  /// набор уровней
  std::vector<Level> levels;
  /// пройденные шаги
  Index steps;
  /// состояние змеи
  Snake snake;
  // Осторожно: конструктор принимает field, поэтому snake дб расположена
  // в классе ниже, чем field, чтобы конструктор field отработал до конструктора snake.
};
