﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Screen.cpp
 *
 * Реализация элементов Screen.hpp.
 */
#include <GL/freeglut.h>
#include "Screen.hpp"

#include <algorithm>
#include <mutex>


//
// рисование текста

/// определить количество строк в тексте (сколько строчек по вертикали будет занимать изображение)
inline std::size_t lines_in_text(const std::string &text)
{
  return 1 + std::count(text.begin(), text.end(), '\n');
}

void drawString(void *font, const std::string &text,
                float left, float bottom, float right, float top)
{
  // функции, вычисляющие "размеры" (в пикселях), не задействуют текущую матрицу,
  // а просто берут значения из зашитых внутри FreeGlut массивов
  auto cstr = reinterpret_cast<const unsigned char*>(text.c_str());
  const GLfloat wdth = static_cast<GLfloat>(glutStrokeLength(font, cstr)), // ширина области рисования
                hght = glutStrokeHeight(font) * lines_in_text(text);       // высота области рисования

  // теперь желательно вписать эти размеры в заданный прямоугольник
  glPushMatrix();
  glTranslatef(left, bottom, 0.0f);
  const GLfloat scale_x = left < right? (right - left) / wdth : 1.0f,
                scale_y = bottom < top? (top - bottom) / hght : 1.0f;

  if (scale_x != 1.0f || scale_y != 1.0f)
    glScalef(scale_x, scale_y, 1.0f);

  // отрисовка текста
  glutStrokeString(font, cstr);
  // восстановление текущей матрицы
  glPopMatrix();
}


//
// подгонка размеров окна

void fitWindow(float target_width, float target_height)
{
  const float scr_width   = float(DrawingAreaInfo::screen_width()),
              scr_height  = DrawingAreaInfo::screen_height() * 0.92f; 
  // магическое число-поправка, рассчитана на горизонтальную панель задач
  // наверняка где-нибудь будет мешать...

  if (scr_width < target_width)
    target_height *= scr_width / target_width, 
    target_width = scr_width;
  if (scr_height < target_height)
    target_width *= scr_height / target_height,
    target_height = scr_height;

  glutPositionWindow(static_cast<int>(0.5f * (scr_width - target_width)),
                     static_cast<int>(0.5f * (scr_height - target_height)));
  glutReshapeWindow(static_cast<int>(target_width), static_cast<int>(target_height));
  // возможно, нужно вводить поправку на ширину границ
}


//
// получение размеров области рисования окна, экрана (в пикселях)

int DrawingAreaInfo::x()
{
  return glutGet(GLUT_WINDOW_X);
}

int DrawingAreaInfo::y()
{
  return glutGet(GLUT_WINDOW_Y);
}

int DrawingAreaInfo::width()
{
  return glutGet(GLUT_WINDOW_WIDTH);
}

int DrawingAreaInfo::height()
{
  return glutGet(GLUT_WINDOW_HEIGHT);
}

int DrawingAreaInfo::border_width()
{
  return glutGet(GLUT_WINDOW_BORDER_WIDTH);
}

int DrawingAreaInfo::border_height()
{
  return glutGet(GLUT_WINDOW_BORDER_HEIGHT);
}

int DrawingAreaInfo::screen_width()
{
  return glutGet(GLUT_SCREEN_WIDTH);
}

int DrawingAreaInfo::screen_height()
{
  return glutGet(GLUT_SCREEN_HEIGHT);
}


//
// AScreen

namespace
{

// текущий экран -- эта глобальная переменная и есть причина невозможности
// переключение экранов из разных параллельно исполняющихся потоков
AScreen *curScreen;
std::mutex curScreen_mut; // на случай попыток переключения экрана из разных нитей одновременно

}

using Lock = std::lock_guard<std::mutex>;

bool AScreen::has_active()
{
  Lock lm(curScreen_mut);
  return curScreen != nullptr;
}

AScreen& AScreen::get_active()
{
  Lock lm(curScreen_mut);
  return *curScreen;
}


AScreen::~AScreen()
{
  Lock lm(curScreen_mut);
  if (curScreen == this)
    curScreenSelect(nullptr);
}


void AScreen::select()
{
  Lock lm(curScreen_mut);
  if (curScreen != this)
  {
    // отключить
    if (curScreen)         
      curScreen->onDeselect();

    curScreenSelect(this);
  }
}


void AScreen::curScreenSelect(AScreen *newScreen)
{
  if (newScreen == curScreen)
    return;

  if (!newScreen) // => curScreen != nullptr
  {
    // отвязать обработчики
    glutIdleFunc          (nullptr);
    glutDisplayFunc       (nullptr);
    glutReshapeFunc       (nullptr);
    glutVisibilityFunc    (nullptr);
    glutKeyboardFunc      (nullptr);
    glutKeyboardUpFunc    (nullptr);
    glutSpecialFunc       (nullptr);
    glutSpecialUpFunc     (nullptr);
    glutEntryFunc         (nullptr);
    glutPassiveMotionFunc (nullptr);
    glutMotionFunc        (nullptr);
    glutMouseFunc         (nullptr);
    glutMouseWheelFunc    (nullptr);
  }
  else if (!curScreen) // => newScreen != nullptr
  {
    // привязать обработчики
    glutIdleFunc          (&AScreen::onIdleWrapper);
    glutDisplayFunc       (&AScreen::onDisplayWrapper);
    glutReshapeFunc       (&AScreen::onReshapeWrapper);
    glutVisibilityFunc    (&AScreen::onVisibilityWrapper);
    glutKeyboardFunc      (&AScreen::onCharWrapper);
    glutKeyboardUpFunc    (&AScreen::onCharUpWrapper);
    glutSpecialFunc       (&AScreen::onSpecialKeyWrapper);
    glutSpecialUpFunc     (&AScreen::onSpecialKeyUpWrapper);
    glutEntryFunc         (&AScreen::onEntryWrapper);
    glutPassiveMotionFunc (&AScreen::onHoverWrapper);
    glutMotionFunc        (&AScreen::onDragWrapper);
    glutMouseFunc         (&AScreen::onMouseButtonWrapper);
    glutMouseWheelFunc    (&AScreen::onMouseWheelWrapper);
  }

  curScreen = newScreen;
  // подключить
  if (curScreen)
  {
    curScreen->onSelect();
    glutShowWindow();
  }
}

// готовые обработчики

void AScreen::onReshape(int width, int height)
{
  glViewport(0, 0, width, height);
}

void AScreen::onDisplayBefore()
{
  glClear(GL_COLOR_BUFFER_BIT); // сброс буфера цвета
}

void AScreen::onDisplayAfter()
{
  glutSwapBuffers(); // обмен буфера
}

// обёртки обработчиков событий -- в основном занимаются простой переадресацией

void AScreen::onIdleWrapper()
{
  curScreen->onIdle();
}

void AScreen::onDisplayWrapper()
{
  curScreen->onDisplayBefore();
  curScreen->onDisplay();
  glFinish();
  curScreen->onDisplayAfter();
}

void AScreen::onReshapeWrapper(int width, int height)
{
  curScreen->onReshape(width, height);
}

void AScreen::onPositionWrapper(int x, int y)
{
  curScreen->onPosition(x, y);
}

void AScreen::onVisibilityWrapper(int status)
{
  curScreen->onVisibility(status == GLUT_VISIBLE);
}

void AScreen::onCharWrapper(unsigned char code, int mouse_x, int mouse_y)
{
  curScreen->onChar(code, mouse_x, mouse_y);
}

void AScreen::onCharUpWrapper(unsigned char code, int mouse_x, int mouse_y)
{
  curScreen->onCharUp(code, mouse_x, mouse_y);
}

void AScreen::onSpecialKeyWrapper(int code, int mouse_x, int mouse_y)
{
  curScreen->onSpecialKey(code, mouse_x, mouse_y);
}

void AScreen::onSpecialKeyUpWrapper(int code, int mouse_x, int mouse_y)
{
  curScreen->onSpecialKeyUp(code, mouse_x, mouse_y);
}

void AScreen::onEntryWrapper(int status)
{
  curScreen->onEntry(status == GLUT_ENTERED);
}

void AScreen::onHoverWrapper(int mouse_x, int mouse_y)
{
  curScreen->onHover(mouse_x, mouse_y);
}

void AScreen::onDragWrapper(int mouse_x, int mouse_y)
{
  curScreen->onDrag(mouse_x, mouse_y);
}

void AScreen::onMouseButtonWrapper(int button, int state, int x, int y)
{
  if (state == GLUT_DOWN)
    curScreen->onMouse(button, x, y);
  else
    curScreen->onMouseUp(button, x, y);
}

void AScreen::onMouseWheelWrapper(int wheel, int direction, int x, int y)
{
  curScreen->onMouseWheel(wheel, direction, x, y);
}
